from pathlib import Path
import re
from zipfile import BadZipFile, is_zipfile, ZipFile

from rpmlint.checks.AbstractCheck import AbstractCheck


class ZipCheck(AbstractCheck):
    """
    Validate zip and jar files correctness.
    """
    zip_regex = re.compile(r'\.(zip|[ewj]ar)$')
    jar_regex = re.compile(r'\.[ewj]ar$')

    def check(self, pkg):
        for fname, pkgfile in pkg.files.items():
            path = pkgfile.path
            if self.zip_regex.search(fname) and Path(path).exists() and \
               Path(path).is_file() and is_zipfile(path):
                try:
                    with ZipFile(path, 'r') as z:
                        # zip checks
                        self._check_bad_crc(pkg, fname, z)
                        self._check_compression(pkg, fname, z)

                        # jar checks
                        if self.jar_regex.search(fname):
                            self._check_classpath(pkg, fname, z)
                            self._check_jarindex(pkg, fname, z)
                except BadZipFile as err:
                    self.output.add_info('E', pkg, 'unable-to-read-zip', f'{fname}: {err}')
                except RuntimeError as err:
                    self.output.add_info('W', pkg, 'unable-to-read-zip', f'{fname}: {err}')

    def _check_bad_crc(self, pkg, fname, zipfile):
        """
        Check CRC issues for the files in the zipfile.

        Print an error if there is a file in the archive that fails CRC check.
        """
        badcrc = zipfile.testzip()
        if badcrc:
            self.output.add_info('E', pkg, 'bad-crc-in-zip', badcrc, fname)

    def _check_compression(self, pkg, fname, zipfile):
        """
        Check if zip is actually compressed or not.

        One file with smaller size is enough. Print an error if the zipfile
        is not compressed.
        """
        # check for empty archives which are valid
        filecount = len(zipfile.namelist())
        nullcount = 0
        for zinfo in zipfile.infolist():
            if zinfo.file_size == 0:
                nullcount += 1
            if zinfo.compress_size != zinfo.file_size:
                return

        # empty files only
        if filecount == nullcount:
            return
        self.output.add_info('E', pkg, 'uncompressed-zip', fname)

    def _check_classpath(self, pkg, fname, jarfile):
        """
        Check if META-INF/MANIFEST.MF file in the jar contains a hardcoded
        Class-Path.

        Print a warning if the path is hardcoded.
        """
        classpath_regex = re.compile(r'^\s*Class-Path\s*:', re.MULTILINE | re.IGNORECASE)

        # the META-INF is optional so skip if it is not present
        mf = 'META-INF/MANIFEST.MF'
        if mf not in jarfile.namelist():
            return

        # otherwise check for the hardcoded classpath
        manifest = jarfile.read(mf).decode(errors='replace')
        if classpath_regex.search(manifest):
            self.output.add_info('W', pkg, 'class-path-in-manifest', fname)

    def _check_jarindex(self, pkg, fname, jarfile):
        """
        Check if the .jar file is indexed.

        Print a warning if 'META-INF/INDEX.LIST' file is not present in the
        jarfile.
        """
        index = 'META-INF/INDEX.LIST'
        if index not in jarfile.namelist():
            self.output.add_info('W', pkg, 'jar-not-indexed', fname)



from Testing import get_tested_mock_package


XCB_PC = """
prefix=/usr
exec_prefix=/usr
libdir=/var/tmp/usr/lib64
includedir=/usr/include//xyz

Name: XCB Xinerama
Description: XCB Xinerama Extension
Version: 1.13
Requires.private: xcb
Libs: -L/usr/lib
Cflags: -I${includedir}
"""


PCPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/tmp/pkgconfig/xcb.pc': {'content': XCB_PC},
    },
    header={'arch': 'x86_64'},
)


LibReiserFSCoreDevelPackage = get_tested_mock_package(
    name='libreiserfscore-devel',
    lazyload=True,
    files={
        '/usr/include/reiserfs/io.h': {
            'content-path': 'files/reiserfs/io.h',
        },
        '/usr/include/reiserfs/misc.h': {
            'content-path': 'files/reiserfs/misc.h',
        },
        '/usr/include/reiserfs/reiserfs_err.h': {
            'content-path': 'files/reiserfs/reiserfs_err.h',
        },
        '/usr/include/reiserfs/reiserfs_fs.h': {
            'content-path': 'files/reiserfs/reiserfs_fs.h',
        },
        '/usr/include/reiserfs/reiserfs_lib.h': {
            'content-path': 'files/reiserfs/reiserfs_lib.h',
        },
        '/usr/include/reiserfs/swab.h': {
            'content-path': 'files/reiserfs/swab.h',
        },
        '/usr/lib64/libreiserfscore.a': {
            'content-path': 'files/reiserfs/libreiserfscore.a',
        },
        '/usr/lib64/libreiserfscore.la': {
            'content-path': 'files/reiserfs/libreiserfscore.la',
        },
        '/usr/lib64/libreiserfscore.so': {
            'link-to': 'libreiserfscore.so.0.0.0',
        },
        '/usr/lib64/libreiserfscore.so.0': {
            'link-to': 'libreiserfscore.so.0.0.0',
        },
        '/usr/lib64/pkgconfig/reiserfscore.pc': {
            'content-path': 'files/reiserfs/reiserfscore.pc',
        },
    },
    header={
        'requires': [
            '/usr/bin/pkg-config',
            'libcom_err-devel',
            'libreiserfscore0 = 3.6.27',
            'libuuid-devel',
            'rpmlib(CompressedFileNames) <= 3.0.4-1',
            'rpmlib(FileDigests) <= 4.6.0-1',
            'rpmlib(PayloadFilesHavePrefix) <= 4.0-1',
            'rpmlib(PayloadIsXz) <= 5.2-1',
        ],
    },
)



from mockdata.mock_pkgconfig import LibReiserFSCoreDevelPackage
import pytest
from rpmlint.checks.BinariesCheck import BinariesCheck
from rpmlint.filter import Filter

from Testing import CONFIG, Config, get_tested_mock_package, get_tested_package, IS_X86_64, TEST_CONFIG


@pytest.fixture(scope='function', autouse=True)
def binariescheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = BinariesCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', ['binary/crypto-policy'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_forbidden_c_calls(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'crypto-policy-non-compliance-openssl /usr/lib/cyrus-imapd/arbitron SSL_CTX_set_cipher_list' in out
    assert 'crypto-policy-non-compliance-openssl /usr/lib64/dovecot/libssl_iostream_openssl.so SSL_CTX_set_cipher_list' in out


@pytest.mark.parametrize('package', ['binary/ngircd'])
def test_waived_forbidden_c_calls(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'crypto-policy-non-compliance' not in out


@pytest.mark.parametrize('package', [LibReiserFSCoreDevelPackage])
def test_lto_bytecode(package, binariescheck):
    output, test = binariescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'lto-bytecode' in out


@pytest.mark.parametrize('package', ['binary/lto-text'])
def test_lto_archive_text(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'lto-no-text-in-archive /usr/lib64/libiberty.a' in out
    assert 'lto-no-text-in-archive /usr/lib64/libdl_p.a' not in out


@pytest.mark.parametrize('package', ['binary/ghc'])
def test_lto_ghc_archive(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'lto-no-text-in-archive' not in out


@pytest.mark.parametrize('package', ['binary/libtool-wrapper'])
def test_libtool_wrapper(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: libtool-wrapper-in-package' in out
    assert 'W: unstripped-binary-or-object' in out
    assert 'E: arch-dependent-file-in-usr-share' in out
    assert 'W: unstripped-binary-or-object /bin/main' in out
    assert 'W: position-independent-executable-suggested /usr/share/main' in out


@pytest.mark.parametrize('package', ['binary/noarch'])
def test_no_arch_issues(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: arch-independent-package-contains-binary-or-object /bin/main' in out
    assert 'E: noarch-with-lib64' in out


@pytest.mark.parametrize('package', [
    get_tested_mock_package(
        lazyload=True,
        files={'/opt/ebpf.o': {'content-path': 'files/ebpf.o', 'create_dirs': True}},
        header={'ARCH': 'noarch'},
    ),
    get_tested_mock_package(
        lazyload=True,
        files={'/opt/python-flit-metadata.txt': {'content-path': 'files/python-flit-metadata.txt', 'create_dirs': True}},
        header={'ARCH': 'noarch'},
    ),
    get_tested_mock_package(
        lazyload=True,
        files={'/opt/x86_64.o': {'content-path': 'files/x86_64.o', 'create_dirs': True}},
        header={'ARCH': 'x86-64'},
    ),
    get_tested_mock_package(
        lazyload=True,
        files={'/opt/aarch64.o': {'content-path': 'files/aarch64.o', 'create_dirs': True}},
        header={'ARCH': 'aarch64'},
    ),
])
def test_no_arch_eBPF(package, binariescheck):
    output, test = binariescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: arch-independent-package-contains-binary-or-object' not in out


@pytest.mark.parametrize('package', [
    get_tested_mock_package(
        lazyload=True,
        files={'/opt/x86_64.o': {'content-path': 'files/x86_64.o', 'create_dirs': True}},
        header={'ARCH': 'noarch'},
    ),
    get_tested_mock_package(
        lazyload=True,
        files={'/opt/aarch64.o': {'content-path': 'files/aarch64.o', 'create_dirs': True}},
        header={'ARCH': 'noarch'},
    ),
])
def test_no_arch_error(package, binariescheck):
    output, test = binariescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: arch-independent-package-contains-binary-or-object' in out


@pytest.mark.parametrize('package', ['binary/libnoexec'])
def test_shlib_with_no_exec(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: shared-library-not-executable /lib64/libfoo.so' not in out


@pytest.mark.parametrize('package', ['binary/glibc'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_shlib_with_no_exec_glibc(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: shared-library-not-executable /lib64/libpthread.so' in out
    assert 'missing-hash-section' not in out
    assert 'missing-gnu-hash-section' not in out


@pytest.mark.parametrize('package', ['binary/bcc-lua'])
def test_position_independent_executable(tmp_path, package, binariescheck):
    CONFIG.configuration['PieExecutables'] = ['.*']
    output = Filter(CONFIG)
    test = BinariesCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: non-position-independent-executable /usr/bin/bcc-lua' in out


@pytest.mark.parametrize('package', ['binary/only-non-binary-in-usr-lib'])
def test_only_non_binary_in_usr_lib(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'W: only-non-binary-in-usr-lib' in out
    # there is a file in /usr/lib64, so no error
    assert 'E: no-binary' not in out
    # we have no 'noarch' or wrapper here
    assert 'E: noarch-with-lib64' not in out
    assert 'E: arch-independent-package-contains-binary-or-object' not in out
    assert 'E: libtool-wrapper-in-package' not in out


# In general we want to throw a warning if we have only non-binary files in
# the /usr/lib. But we can allow non-binaries via UsrLibBinaryException config
# option. These files will be considered binaries and no warning should be
# thrown.
@pytest.mark.parametrize('package',
                         ['binary/only-non-binary-in-usr-lib_exception'])
def test_only_non_binary_in_usr_lib_exception(tmp_path, package, binariescheck):
    config = Config(TEST_CONFIG)
    config.configuration['UsrLibBinaryException'] = '^/usr/lib(64)?/python'
    output = Filter(config)
    test = BinariesCheck(config, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'W: only-non-binary-in-usr-lib' not in out


@pytest.mark.parametrize('package', ['binary/no-binary'])
def test_no_binary(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: no-binary' in out
    # no .la file or binary there
    assert 'E: invalid-la-file' not in out
    assert 'E: binary-in-etc' not in out


@pytest.mark.parametrize('package', ['binary/invalid-la-file'])
def test_invalid_la_file(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: invalid-la-file' in out
    # no /usr/share dir there
    assert 'E: arch-dependent-file-in-usr-share' not in out


@pytest.mark.parametrize('package', ['binary/binary-in-etc'])
def test_binary_in_etc(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: binary-in-etc' in out
    # it's not a library package
    assert 'E: executable-in-library-package' not in out


@pytest.mark.parametrize('package', ['binary/non-position-independent-exec'])
def test_non_position_independent_sugg(tmp_path, package, binariescheck):
    # reset PieExecutable option
    CONFIG.configuration['PieExecutables'] = []
    output = Filter(CONFIG)
    test = BinariesCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'W: position-independent-executable-suggested' in out
    # it should throw just a warning as it's not forced by PieExecutables opt
    assert 'E: non-position-independent-executable' not in out


# Force an error by setting PieExecutables option to the no-pie binary
@pytest.mark.parametrize('package', ['binary/non-position-independent-exec'])
def test_non_position_independent(tmp_path, package, binariescheck):
    CONFIG.configuration['PieExecutables'] = ['sparta', '.*hello']
    output = Filter(CONFIG)
    test = BinariesCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: non-position-independent-executable' in out
    # It should throw just the error, not warning
    assert 'W: position-independent-executable-suggested' not in out


# libtest package
@pytest.mark.parametrize('package', ['binary/libtest'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_library(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: executable-in-library-package' in out
    assert 'W: no-soname' in out
    # there is no soname here so it can't be invalid
    assert 'E: invalid-soname' not in out


# invalid-soname test package
@pytest.mark.parametrize('package', ['binary/libtest1'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_shared_library1(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: invalid-soname' in out
    # there is an invalid soname here, so no "no-soname" error
    assert 'W: no-soname' not in out


# shlib-policy-name-error test package
@pytest.mark.parametrize('package', ['binary/libtest2'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_shared_library2(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: shlib-policy-name-error' in out
    # it doesn't call /sbin/ldconfig
    assert 'E: no-ldconfig-symlink' in out
    # no ldconfig is not invalid
    assert 'E: invalid-ldconfig-symlink' not in out
    # the soname is set
    assert 'W: no-soname' not in out


# invalid-ldconfig-symlink test package
@pytest.mark.parametrize('package', ['binary/libtest3'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_invalid_ldconfig_symlink(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: invalid-ldconfig-symlink' in out
    # executable doesn't call mktemp, setuid or gethostbyname
    assert 'E: call-to-mktemp' not in out
    assert 'E: missing-call-to-setgroups-before-setuid' not in out
    assert 'W: binary-or-shlib-calls-gethostbyname' not in out
    # it's not statically linked either
    assert 'E: statically-linked-binary' not in out


# valid symlink should not report invalid-ldconfig-symlink
@pytest.mark.parametrize('package', ['binary/libtest4'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_not_valid_ldconfig_symlink(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: invalid-ldconfig-symlink' not in out


@pytest.mark.parametrize('package', ['binary/multiple_errors'])
@pytest.mark.skipif(not IS_X86_64, reason='x86-64 only')
def test_multiple_errors(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: call-to-mktemp' in out
    assert 'E: missing-call-to-setgroups-before-setuid' in out
    assert 'W: binary-or-shlib-calls-gethostbyname' in out


@pytest.mark.parametrize('package', ['binary/libtest'])
def test_patchable_function_entry_archive(tmp_path, package, binariescheck):
    output, test = binariescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: patchable-function-entry-in-archive /usr/lib64/libhello.a' in out


@pytest.mark.parametrize('package', [
    get_tested_mock_package(files=['/usr/lib/systemd/system/yast-timesync.service']),
])
def test_systemd_unit_file(package, binariescheck):
    output, test = binariescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'only-non-binary-in-usr-lib' not in out



from Testing import get_tested_mock_package


########################
# Python mock packages #
########################

# Not valid doc folder in python packages
PythonDocFolderPackage = get_tested_mock_package(
    lazyload=True,
    files=[
        '/usr/lib/python2.7/site-packages/python-mypackage/doc',
        '/usr/lib/python2.7/site-packages/python-mypackage/docs',
        '/usr/lib/python3.10/site-packages/python-mypackage/doc',
        '/usr/lib/python3.10/site-packages/python-mypackage/docs',
        '/usr/lib64/python2.7/site-packages/python-mypackage/doc',
        '/usr/lib64/python2.7/site-packages/python-mypackage/docs',
        '/usr/lib64/python3.10/site-packages/python-mypackage/doc',
        '/usr/lib64/python3.10/site-packages/python-mypackage/docs',
    ]
)


# Valid doc folder in python package if it's a python module
PythonDocModulePackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python2.7/site-packages/python-mypackage/doc/__init__.py': {'create_dirs': True, 'include_dirs': 2},
        '/usr/lib/python2.7/site-packages/python-mypackage/docs/__init__.py': {'create_dirs': True, 'include_dirs': 1},
        '/usr/lib64/python2.7/site-packages/python-mypackage/doc/__init__.py': {'create_dirs': True, 'include_dirs': 2},
        '/usr/lib64/python2.7/site-packages/python-mypackage/docs/__init__.py': {'create_dirs': True, 'include_dirs': 1},
        '/usr/lib/python3.10/site-packages/python-mypackage/doc/__init__.py': {'create_dirs': True, 'include_dirs': 2},
        '/usr/lib/python3.10/site-packages/python-mypackage/docs/__init__.py': {'create_dirs': True, 'include_dirs': 1},
        '/usr/lib64/python3.10/site-packages/python-mypackage/doc/__init__.py': {'create_dirs': True, 'include_dirs': 2},
        '/usr/lib64/python3.10/site-packages/python-mypackage/docs/__init__.py': {'create_dirs': True, 'include_dirs': 1},
    }
)


# Python package with old egginfo file, not folder
PythonEggInfoFileackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python2.7/site-packages/mydistutilspackage.egg-info': {'content': 'Metadata-Version: 2.1\nName: pythoncheck'},
        '/usr/lib/python3.10/site-packages/mydistutilspackage.egg-info': {'content': 'Metadata-Version: 2.1\nName: pythoncheck'},
        '/usr/lib64/python2.7/site-packages/mydistutilspackage.egg-info': {'content': 'Metadata-Version: 2.1\nName: pythoncheck'},
        '/usr/lib64/python3.10/site-packages/mydistutilspackage.egg-info': {'content': 'Metadata-Version: 2.1\nName: pythoncheck'},
    }
)


PythonFlitMockPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.10/site-packages/flit-3.8.0.dist-info/METADATA': {
            'content-path': 'files/python-flit-metadata.txt',
            'create_dirs': True,
        },
    },
    header={
        'requires': [
            'python-flit_core',
            'python-requests',
            'python-tomli-w',
            'python310-docutils',
        ],
    },
)


# Python package with missing require
PythonFlitMissingRequirePackage = PythonFlitMockPackage.clone(
    header={
        'requires': [
            'python3-flit-core',
            'python3-requests',
            'python3-tomli-w',
        ],
    },
)


# Python package with missing require
PythonFlitLeftoverRequirePackage = PythonFlitMockPackage.clone(
    header={
        'requires': [
            'python3-docutils',
            'python3-flit-core',
            'python3-poetry',
            'python3-requests',
            'python3-tomli-w',
        ],
    },
)


PythonJupyterServerFileidMockPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.10/site-packages/jupyter_server_fileid-0.9.0.dist-info/METADATA': {
            'content-path': 'files/python-jupyter_server_fileid-metadata.txt',
            'create_dirs': True,
        },
    },
    header={
        'requires': [
            'python-jupyter-events',
            'python-click',
            'python-jupyter-server',
        ],
    },
)


PythonJupyterEventsMockPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.10/site-packages/jupyter_events-0.6.3.dist-info/METADATA': {
            'content-path': 'files/python-jupyter-events-metadata.txt',
            'create_dirs': True,
        },
    },
    header={
        'requires': [
            'python-jsonschema',
            'python-python-json-logger',
            'python-pyyaml',
            'python-rfc3339-validator',
            'python-rfc3986-validator',
            'python-traitlets',
            'python-click',
            'python-rich',
            'python-jupyterlite-sphinx',
            'python-myst-parser',
            'python-pydata-sphinx-theme',
            'python-sphinxcontrib-spelling',
            'python-click',
            'python-coverage',
            'python-pre-commit',
            'python-pytest-asyncio',
            'python-pytest-console-scripts',
            'python-pytest-cov',
            'python-pytest',
            'python-rich',
        ],
    },
)


PythonScikitBuildMockPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.10/site-packages/scikit_build-0.17.2.dist-info/METADATA': {
            'content-path': 'files/python-scikit_build-metadata.txt',
            'create_dirs': True,
        },
    },
    header={
        'requires': [
            'python-distro',
            'python-packaging',
            'python-setuptools',
            'python-wheel',
            'python-tomli',
        ],
    },
)


# Python flit package with python3.12dist(foo) requirements (used in Fedora see #1171)
PythonFlitFedoraMockPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.10/site-packages/flit-3.8.0.dist-info/METADATA': {
            'content-path': 'files/python-flit-metadata.txt',
            'create_dirs': True,
        },
    },
    header={
        'requires': [
            'python3.12dist(docutils)',
            'python3.12dist(flit-core) >= 3.8',
            'python3.12dist(requests)',
            'python3.12dist(tomli-w)',
        ],
    },
)


# Python package with old setuptools metadata format, egg-info folder
PythonIcecreamPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.10/site-packages/icecream-2.1.3-py3.10.egg-info/requires.txt': {
            'content': """
asttokens>=2.0.1
colorama>=0.3.9
executing>=0.3.1
pygments>=2.2.0
""",
            'create_dirs': True
        },
    },
    header={
        'requires': [
            'asttokens>=2.0.1',
            'colorama>=0.3.9',
            'executing>=0.3.1',
            'pygments>=2.2.0',
        ],
    },
)


# Python package with missing require
PythonIcecreamMissingRequirePackage = PythonIcecreamPackage.clone(
    header={
        'requires': [
            'asttokens>=2.0.1',
            'executing>=0.3.1',
            'pygments>=2.2.0',
        ],
    },
)


# Python package with leftover require
PythonIcecreamLeftoverRequirePackage = PythonIcecreamPackage.clone(
    header={
        'requires': [
            'python3-asttokens >= 2.0.1',
            'python3-colorama >= 0.3.9',
            'python3-executing >= 0.3.1',
            'python3-poetry',
            'python3-pygments >= 2.2.0',
        ],
    },
)


# Python package with multiple pyc for different python versions in the same
# sitelib
PythonMultiplePYCMockPackage = get_tested_mock_package(
    lazyload=True,
    files=[
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/base.cpython-310.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/base.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/base.cpython-39.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/__init__.cpython-310.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/__init__.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/__init__.cpython-39.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_saferef.cpython-310.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_saferef.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_saferef.cpython-39.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_utilities.cpython-310.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_utilities.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_utilities.cpython-39.pyc',
    ]
)


# Python package with pyc files but just the correct version
PythonSinglePYCMockPackage = get_tested_mock_package(
    lazyload=True,
    files=[
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/base.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/base.cpython-39.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/__init__.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/__init__.cpython-39.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_saferef.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_saferef.cpython-39.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_utilities.cpython-39.opt-1.pyc',
        '/usr/lib/python3.9/site-packages/blinker/__pycache__/_utilities.cpython-39.pyc',
    ]
)


IPythonMissingRequirePackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/python3.12/site-packages/ipython-8.14.0-py3.12.egg-info/requires.txt': {
            'content-path': 'files/ipython-requires.txt',
        },
    },
    header={'requires': [
        'python-leftover',
        'python-no-leftover',
    ]},
)



from mockdata.mock_LSB import NonLSBCompliancePackage
import pytest
from rpmlint.checks.LSBCheck import LSBCheck
from rpmlint.filter import Filter

from Testing import CONFIG


@pytest.fixture(scope='function', autouse=True)
def lsbcheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = LSBCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', [NonLSBCompliancePackage])
def test_LSB_compliance(package, lsbcheck):
    """
    Check that the package name, version and release number are LSB compliant.
    """
    output, test = lsbcheck
    test.check(package)
    out = output.print_results(output.results)

    # Check invalid package name
    assert 'non-lsb-compliant-package-name' in out
    assert 'package name contains an illegal character' in out

    # Check invalid package version
    assert 'non-lsb-compliant-version' in out
    assert 'version number contains an illegal character' in out

    # Check invalid package release
    assert 'non-lsb-compliant-release' in out
    assert 'release number contains an illegal character' in out



from Testing import get_tested_mock_package


BuildDatePackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/bin/with-date': {'content': 'Jan  1 2019'},
        '/bin/with-datetime': {'content': 'Jan  1 2019 12:15:11'},
    },
)



from pathlib import Path

from rpmlint.config import Config
from rpmlint.filter import Filter

from Testing import get_tested_package, get_tested_path

TEST_CONFIG_FILTERS = [get_tested_path('configs/testfilters.config')]
TEST_RPMLINTRC = get_tested_path('configs/testing-rpmlintrc')
TEST3_RPMLINTRC = get_tested_path('configs/testing3-rpmlintrc')
TEST_PACKAGE = Path('binary', 'ngircd')
TEST_PACKAGE2 = Path('binary', 'tempfiled')
TEST_DESCRIPTIONS = [get_tested_path('configs/descriptions.config')]


def test_filters_regexp():
    """
    Load some filters and make sure we generate nice regexp
    """
    cfg = Config(TEST_CONFIG_FILTERS)
    assert len(cfg.configuration['Filters']) == 12
    assert cfg.configuration['Filters'][0] == '.*invalid-buildhost.*'


def test_data_storing(tmp_path):
    """
    Load some filters and make sure we generate nice regexp
    """
    cfg = Config(TEST_CONFIG_FILTERS)
    cfg.load_rpmlintrc(TEST_RPMLINTRC)
    result = Filter(cfg)
    pkg = get_tested_package(TEST_PACKAGE, tmp_path)
    # this should be upgraded to error
    result.add_info('I', pkg, 'suse-other-error', '')
    assert len(result.results) == 1
    assert result.printed_messages['I'] == 0
    assert result.printed_messages['E'] == 1
    # this should be downgraded
    result.add_info('E', pkg, 'suse-dbus-unauthorized-service', '')
    assert len(result.results) == 2
    assert result.printed_messages['W'] == 1
    assert result.printed_messages['E'] == 1


def test_data_storing_backward_compat(tmp_path):
    """
    Make sure we can load some filters from rpmlintrc that
    worked with rpmlint v1.
    """
    cfg = Config(TEST_CONFIG_FILTERS)
    cfg.load_rpmlintrc(TEST3_RPMLINTRC)
    parsed_filters = cfg.rpmlintrc_filters
    assert 'no-spaces-in-paren' in parsed_filters
    assert 'has-spaces-in-paren' in parsed_filters
    assert 'multiple-spaces-in-paren' in parsed_filters
    assert 'doublequotes-instead-of-singlequotes' in parsed_filters


def test_description_storing(tmp_path):
    """
    Test if we can store extra destcriptions and formatting is up par
    """
    lorem_formated = """Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore
eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt
in culpa qui officia deserunt mollit anim id est laborum.\n\n"""
    cfg = Config(TEST_CONFIG_FILTERS)
    result = Filter(cfg)
    pkg = get_tested_package(TEST_PACKAGE, tmp_path)
    assert len(result.results) == 0
    result.add_info('E', pkg, 'suse-dbus-unauthorized-service', '')
    # two options so we check the description is added only once
    result.add_info('I', pkg, 'suse-other-error', '/usr/bin/1')
    # nothing is populated
    assert not result.get_description('suse-other-error')
    # add descriptions
    result.error_details.update({'suse-other-error': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.'})
    assert result.get_description('suse-other-error') == lorem_formated


def test_description_from_toml(tmp_path):
    """
    Test if description loaded from toml shows up details
    """
    cfg = Config(TEST_CONFIG_FILTERS)
    result = Filter(cfg)
    assert result.get_description('uncompressed-zip')
    assert result.get_description('uncompressed-zip') == 'The zip file is not compressed.\n\n'


def test_description_from_conf(tmp_path):
    """
    Test that descriptions strings are updated from configuration file.

    Load [Descriptions] from TEST_DESCRIPTIONS config file and test that the
    rpmlint error details are updated to the new values.
    """
    cfg = Config(TEST_DESCRIPTIONS)
    result = Filter(cfg)

    assert result.get_description('no-binary', cfg)
    assert result.get_description('no-binary', cfg) == \
        'A new text for no-binary error.\n\n'

    assert result.get_description('no-soname', cfg)
    assert result.get_description('no-soname', cfg) == \
        'A new text for no-soname error.\n\n'

    # At this point, only basic descriptions from "descriptions" directory are
    # loaded. "Dynamic" descriptions that are defined directly in the check
    # file (e.g. see FHSCheck.py and its fhs_details_dict) are loaded later so
    # we can't test it now. It's tested in test_lint.py.

    assert not result.get_description('non-standard-dir-in-usr', cfg)
    assert result.get_description('non-standard-dir-in-usr', cfg) != \
        'A new text for non-standard-dir-in-usr error.\n\n'

    assert not result.get_description('non-standard-dir-in-var', cfg)
    assert result.get_description('non-standard-dir-in-var', cfg) != \
        'A new text for non-standard-dir-in-var error.\n\n'


def test_output(tmp_path):
    """
    Test the actual output of rpmlint on one file
    """
    expected_output = """ngircd.x86_64: I: suse-other-error /usr/bin/1
ngircd.x86_64: I: suse-other-error /usr/bin/2
tempfiled.x86_64: E: suse-other-error /usr/bin/3
Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor
incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis
nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat.
Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore
eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt
in culpa qui officia deserunt mollit anim id est laborum.

ngircd.x86_64: E: suse-dbus-unauthorized-service\n"""
    cfg = Config(TEST_CONFIG_FILTERS)
    result = Filter(cfg)
    pkg = get_tested_package(TEST_PACKAGE, tmp_path)
    pkg2 = get_tested_package(TEST_PACKAGE2, tmp_path)
    # here we check if empty detail will not add whitespace
    result.add_info('E', pkg, 'suse-dbus-unauthorized-service', '')
    # two options so we check the description is added only once
    result.add_info('I', pkg, 'suse-other-error', '/usr/bin/1')
    result.add_info('I', pkg, 'suse-other-error', '/usr/bin/2')
    result.add_info('E', pkg2, 'suse-other-error', '/usr/bin/3')
    result.error_details.update({'suse-other-error': 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.'})
    assert len(result.print_results(result.results).splitlines()) == 4
    result.info = True
    assert len(result.print_results(result.results).splitlines()) == 11
    assert result.print_results(result.results) == expected_output


def test_filtered_output(tmp_path):
    cfg = Config(TEST_CONFIG_FILTERS)
    result = Filter(cfg)
    pkg = get_tested_package(TEST_PACKAGE, tmp_path)
    assert len(result.results) == 0
    result.add_info('E', pkg, 'no-regex', '')
    result.add_info('E', pkg, 'no-regex-with-leading-space', '')
    result.add_info('E', pkg, 'bad-error', 'details of the error')
    result.add_info('E', pkg, 'test-color-error', 'details of the error')
    assert len(result.results) == 0


def test_blocked_filters(tmp_path):
    key = 'fatal-error'
    cfg = Config(TEST_CONFIG_FILTERS)
    result = Filter(cfg)
    pkg = get_tested_package(TEST_PACKAGE, tmp_path)
    assert len(result.results) == 0
    assert key in cfg.configuration['Filters']
    result.add_info('E', pkg, key, '')
    assert len(result.results) == 1



from mockdata.mock_pam_modules import PAMModulePackage
import pytest
from rpmlint.checks.PAMModulesCheck import PAMModulesCheck
from rpmlint.filter import Filter

from Testing import CONFIG


@pytest.fixture(scope='function', autouse=True)
def pammodulecheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = PAMModulesCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', [PAMModulePackage])
def test_pam_modules(package, pammodulecheck):
    output, test = pammodulecheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: pam-unauthorized-module pam-module.so' in out



import re
import stat

from rpmlint.checks.AbstractCheck import AbstractFilesCheck


class PkgConfigCheck(AbstractFilesCheck):
    """
    Validate that .pc files are correct.
    """
    suspicious_dir = re.compile(r'[=:](?:/usr/src/\w+/BUILD|/var/tmp|/tmp|/home)')

    def __init__(self, config, output):
        super().__init__(config, output, r'.*/pkgconfig/.*\.pc$')

    def check(self, pkg):
        # check for references to /lib when in lib64 mode and vice versa
        if pkg.arch in ('x86_64', 'ppc64', 's390x', 'aarch64'):
            self.wronglib_dir = re.compile(r'-L/usr/lib\b')
        else:
            self.wronglib_dir = re.compile(r'-L/usr/lib64\b')

        AbstractFilesCheck.check(self, pkg)

    def check_file(self, pkg, filename):
        if pkg.is_source or not stat.S_ISREG(pkg.files[filename].mode):
            return

        try:
            with open(pkg.dir_name() + '/' + filename, encoding='utf-8') as pc_file:
                for line in pc_file:
                    self._check_invalid_pkgconfig_file(pkg, filename, line)
                    self._check_invalid_libs_dir(pkg, filename, line)
                    self._check_double_slash(pkg, filename, line)
        except Exception as e:
            self.output.add_info('E', pkg, 'pkgconfig-exception', filename, str(e))

    def _check_invalid_pkgconfig_file(self, pkg, filename, line):
        """
        Check that .pc file is valid (it runs various checks).

        E.g. it doesn't contain traces of $RPM_BUILD_ROOT or $RPM_BUILD_DIR,
        unreplaced macros or invalid paths.
        """
        if self.suspicious_dir.search(line):
            self.output.add_info('E', pkg, 'invalid-pkgconfig-file', filename)

    def _check_invalid_libs_dir(self, pkg, filename, line):
        """
        Check that .pc file contains correct libs dir based on the build
        target (32-bit, 64-bit).

        That means:
         -L/usr/lib or -L/lib for 32-bit,
         -L/usr/lib64 or -L/lib64 for 64-bit
        """
        if line.startswith('Libs:') and self.wronglib_dir.search(line):
            self.output.add_info('E', pkg, 'pkgconfig-invalid-libs-dir', filename, line.rstrip())

    def _check_double_slash(self, pkg, filename, line):
        """
        Check that .pc file doesn't contain a path with a double slash ('//')
        """
        if '//' in line and '://' not in line:
            self.output.add_info('E', pkg, 'double-slash-in-pkgconfig-path', filename, line.rstrip())



from itertools import dropwhile, takewhile
import re
import subprocess

from rpmlint.helpers import ENGLISH_ENVIRONMENT


class ElfSection:
    """
    A simple wrapper representing one ELF section.
    """
    def __init__(self, name, size):
        self.name = name
        self.size = int(size, 16)


class ElfProgramHeader:
    """
    A simple wrapper representing one ELF program header.
    """
    def __init__(self, name, flags):
        self.name = name
        self.flags = flags.replace(' ', '')


class ElfDynamicSection:
    """
    A simple wrapper representing one ELF dynamic section entry.
    """
    def __init__(self, key, value):
        self.key = key
        self.value = value


class ElfSectionInfo:
    """
    Class contains information about ELF sections of an ELF file. The information
    is get with the following command line: readelf -WS.

    Output example:

    There are 12 section headers, starting at offset 0x268:

    Section Headers:
      [Nr] Name              Type            Address          Off    Size   ES Flg Lk Inf Al
      [ 0]                   NULL            0000000000000000 000000 000000 00      0   0  0
      [ 1] .text             PROGBITS        0000000000000000 000040 000015 00  AX  0   0  1
      [ 2] .rela.text        RELA            0000000000000000 0001d8 000018 18   I  9   1  8
      [ 3] .data             PROGBITS        0000000000000000 000055 000000 00  WA  0   0  1
      [ 4] .bss              NOBITS          0000000000000000 000055 000000 00  WA  0   0  1
      [ 5] .comment          PROGBITS        0000000000000000 000055 000041 01  MS  0   0  1
      [ 6] .note.GNU-stack   PROGBITS        0000000000000000 000096 000000 00      0   0  1
      [ 7] .eh_frame         PROGBITS        0000000000000000 000098 000038 00   A  0   0  8
      [ 8] .rela.eh_frame    RELA            0000000000000000 0001f0 000018 18   I  9   7  8
      [ 9] .symtab           SYMTAB          0000000000000000 0000d0 0000f0 18     10   8  8
      [10] .strtab           STRTAB          0000000000000000 0001c0 000011 00      0   0  1
      [11] .shstrtab         STRTAB          0000000000000000 000208 000059 00      0   0  1
    Key to Flags:
      W (write), A (alloc), X (execute), M (merge), S (strings), I (info),
      L (link order), O (extra OS processing required), G (group), T (TLS),
      C (compressed), x (unknown), o (OS specific), E (exclude),
      l (large), p (processor specific)
    """

    section_regex = re.compile(r'.*\] (?P<section>\S*)\s*\S+\s*\S*\s*\S*\s*(?P<size>\w*)')
    pic_regex = re.compile(r'\.rela?\.(data|text)')

    def __init__(self, path, extra_flags):
        self.path = path
        self.elf_files = []
        self.parsing_failed_reason = None
        self.pic = False
        self.extra_flags = extra_flags
        self.parse()

    def parse(self):
        r = subprocess.run(['readelf', '-W', '-S', self.path] + self.extra_flags, encoding='utf8',
                           errors='replace', capture_output=True, env=ENGLISH_ENVIRONMENT)
        if r.returncode != 0:
            self.parsing_failed_reason = r.stderr
            return

        lines = r.stdout.splitlines()
        needle = 'Section Headers:'

        # archive files can contain multiple files
        i = 0
        length = len(lines)

        while i < length:
            parsed_sections = []
            while needle not in lines[i]:
                i += 1
                if i == length:
                    return

            # skip header and empty section
            i += 3

            sections = []
            while 'Key to Flags:' not in lines[i]:
                sections.append(lines[i])
                i += 1

            for s in sections:
                r = self.section_regex.search(s)
                section = ElfSection(r.group('section'), r.group('size'))
                parsed_sections.append(section)

                # detect a PIC section
                if self.pic_regex.search(section.name) is not None:
                    self.pic = True

            if len(parsed_sections) > 0:
                self.elf_files.append(parsed_sections)


class ElfProgramHeaderInfo:
    """
    Program Headers:
      Type           Offset   VirtAddr           PhysAddr           FileSiz  MemSiz   Flg Align
      PHDR           0x000040 0x0000000000400040 0x0000000000400040 0x000268 0x000268 R   0x8
      INTERP         0x0002a8 0x00000000004002a8 0x00000000004002a8 0x00001c 0x00001c R   0x1
          [Requesting program interpreter: /lib64/ld-linux-x86-64.so.2]
      LOAD           0x000000 0x0000000000400000 0x0000000000400000 0x000460 0x000460 R   0x1000
      LOAD           0x001000 0x0000000000401000 0x0000000000401000 0x0002ad 0x0002ad R E 0x1000
      LOAD           0x002000 0x0000000000402000 0x0000000000402000 0x0001d0 0x0001d0 R   0x1000
      LOAD           0x002e00 0x0000000000403e00 0x0000000000403e00 0x000230 0x000238 RW  0x1000
      DYNAMIC        0x002e10 0x0000000000403e10 0x0000000000403e10 0x0001e0 0x0001e0 RW  0x8
      NOTE           0x0002c4 0x00000000004002c4 0x00000000004002c4 0x000044 0x000044 R   0x4
      GNU_EH_FRAME   0x002004 0x0000000000402004 0x0000000000402004 0x000054 0x000054 R   0x4
      GNU_STACK      0x000000 0x0000000000000000 0x0000000000000000 0x000000 0x000000 RW  0x10
      GNU_RELRO      0x002e00 0x0000000000403e00 0x0000000000403e00 0x000200 0x000200 R   0x1
    """

    header_regex = re.compile('\\s+(?P<header>\\w+)(\\s+\\w+){5}\\s+(?P<flags>[RWE ]{3}).*')

    def __init__(self, path, extra_flags):
        self.path = path
        self.headers = []
        self.parsing_failed_reason = None
        self.extra_flags = extra_flags
        self.parse()

    def parse(self):
        r = subprocess.run(['readelf', '-W', '-l', self.path] + self.extra_flags, encoding='utf8',
                           errors='replace', capture_output=True, env=ENGLISH_ENVIRONMENT)
        if r.returncode != 0:
            self.parsing_failed_reason = r.stderr
            return

        lines = r.stdout.splitlines()
        needle = 'Program Headers:'

        while len(lines) > 0:
            lines = list(dropwhile(lambda x: needle not in x, lines))

            # skip header
            lines = lines[2:]

            sections = list(takewhile(lambda x: x.strip() != '', lines))
            for s in sections:
                r = self.header_regex.search(s)
                if r is not None:
                    self.headers.append(ElfProgramHeader(r.group('header'), r.group('flags')))

            lines = lines[len(sections):]


class ElfDynamicSectionInfo:
    """
    0x0000000000000001 (NEEDED)             Shared library: [ld-linux-x86-64.so.2]
    0x000000000000000e (SONAME)             Library soname: [libc.so.6]
    0x000000000000000c (INIT)               0x26950
    0x0000000000000019 (INIT_ARRAY)         0x1ba330
    0x000000000000001b (INIT_ARRAYSZ)       16 (bytes)
    0x0000000000000004 (HASH)               0x328
    0x000000006ffffef5 (GNU_HASH)           0x37f8
    0x0000000000000005 (STRTAB)             0x151e0
    0x0000000000000006 (SYMTAB)             0x7488
    0x000000000000000a (STRSZ)              24691 (bytes)
    0x000000000000000b (SYMENT)             24 (bytes)
    0x0000000000000003 (PLTGOT)             0x1bcbd0
    0x0000000000000002 (PLTRELSZ)           1152 (bytes)
    0x0000000000000014 (PLTREL)             RELA
    0x0000000000000017 (JMPREL)             0x24538
    0x0000000000000007 (RELA)               0x1c948
    0x0000000000000008 (RELASZ)             31728 (bytes)
    0x0000000000000009 (RELAENT)            24 (bytes)
    0x000000006ffffffc (VERDEF)             0x1c4c8
    0x000000006ffffffd (VERDEFNUM)          31
    0x000000000000001e (FLAGS)              BIND_NOW STATIC_TLS
    0x000000006ffffffb (FLAGS_1)            Flags: NOW
    0x000000006ffffffe (VERNEED)            0x1c918
    0x000000006fffffff (VERNEEDNUM)         1
    0x000000006ffffff0 (VERSYM)             0x1b254
    0x000000006ffffff9 (RELACOUNT)          1232
    0x0000000000000000 (NULL)               0x0

    handle also:

    0x60009990 (Operating System specific: 60009990)        0x24e20
    0x60009991 (Operating System specific: 60009991)        0x8
    """

    section_regex = re.compile('\\s+\\w*\\s+\\((?P<key>[^\\)]+)\\)\\s+(?P<value>.*)')
    soname_regex = re.compile('Library soname: \\[(?P<soname>[^\\]]+)\\]')
    needed_regex = re.compile('Shared library: \\[(?P<library>[^\\]]+)\\]')
    runpath_regex = re.compile('Library runpath: \\[(?P<path>[^\\]]+)\\]')
    rpath_regex = re.compile('Library rpath: \\[(?P<path>[^\\]]+)\\]')

    def __init__(self, path, extra_flags):
        self.path = path
        self.sections = []
        self.parsing_failed_reason = None
        self.extra_flags = extra_flags
        self.parse()
        self.parse_meta()

    def parse(self):
        r = subprocess.run(['readelf', '-W', '-d', self.path] + self.extra_flags, encoding='utf8',
                           errors='replace', capture_output=True, env=ENGLISH_ENVIRONMENT)
        if r.returncode != 0:
            self.parsing_failed_reason = r.stderr
            return

        lines = r.stdout.splitlines()
        needle = 'Dynamic section at offset'

        lines = list(dropwhile(lambda x: needle not in x, lines))
        # skip header
        lines = lines[2:]
        for line in lines:
            r = self.section_regex.search(line)
            self.sections.append(ElfDynamicSection(r.group('key'), r.group('value')))

    def parse_meta(self):
        self.soname = None
        soname = self['SONAME']
        if len(soname) == 1:
            r = self.soname_regex.search(soname[0])
            if r:
                self.soname = r.group('soname')

        self.needed = []
        for line in self['NEEDED']:
            r = self.needed_regex.search(line)
            if r:
                self.needed.append(r.group('library'))

        self.runpaths = []

        # Parse both RUNPATH and RPATH
        for line in self['RUNPATH']:
            r = self.runpath_regex.search(line)
            if r:
                self.runpaths.append(r.group('path'))
        for line in self['RPATH']:
            r = self.rpath_regex.search(line)
            if r:
                self.runpaths.append(r.group('path'))

    def __getitem__(self, key):
        return [x.value for x in self.sections if x.key == key]


class ElfSymbolTableInfo:
    """
     7: 0000000000000000     0 SECTION LOCAL  DEFAULT    7
     8: 0000000000000000     0 SECTION LOCAL  DEFAULT    8
     9: 0000000000000000     0 SECTION LOCAL  DEFAULT    6
    10: 0000000000000000    18 FUNC    GLOBAL DEFAULT    4 main
    11: 0000000000000000    11 FUNC    GLOBAL DEFAULT    5 foo
    ...
     7: 0000000000000000     0 SECTION LOCAL  DEFAULT    5 .comment
     8: 0000000000000000    21 FUNC    GLOBAL DEFAULT    1 main
    """

    def __init__(self, path, extra_flags):
        self.path = path
        self.functions = set()
        self.parsing_failed_reason = None
        self.extra_flags = extra_flags
        self.parse()

    def parse(self):
        try:
            r = subprocess.run(['readelf', '-Ui', '-W', '-s', self.path] + self.extra_flags, encoding='utf8',
                               errors='replace', capture_output=True, env=ENGLISH_ENVIRONMENT)
            if r.returncode != 0:
                self.parsing_failed_reason = r.stderr
                return

            lines = r.stdout.splitlines()
            for line in lines:
                parts = line.split()
                if len(parts) >= 8 and parts[3] == 'FUNC':
                    self.functions.add(parts[7])
        except UnicodeDecodeError as e:
            self.parsing_failed_reason = str(e)

    def get_functions_for_regex(self, regex):
        for sym in self.functions:
            if regex.search(sym):
                yield sym


class ElfCommentInfo:
    """
    String dump of section '.comment':
      [     1]  GHC 8.6.5
    """

    comment_regex = re.compile('\\s+\\[[\\s[0-9]+\\]\\s+(?P<comment>.*)')

    def __init__(self, path, extra_flags):
        self.path = path
        self.comments = []
        self.parsing_failed_reason = None
        self.extra_flags = extra_flags
        self.parse()

    def parse(self):
        r = subprocess.run(['readelf', '-p', '.comment', self.path] + self.extra_flags, encoding='utf8',
                           errors='replace', capture_output=True, env=ENGLISH_ENVIRONMENT)
        if r.returncode != 0:
            self.parsing_failed_reason = r.stderr
            return

        lines = r.stdout.splitlines()
        for line in lines:
            r = self.comment_regex.search(line)
            if r:
                self.comments.append(r.group('comment'))


class ReadelfParser:
    """
    Class contains all information obtained by readelf command
    in a structured format.
    """

    NOT_ELF_ERROR = 'Error: Not an ELF file - it has the wrong magic bytes at the start'
    so_regex = re.compile(r'/lib(64)?/[^/]+\.so(\.[0-9]+)*$')

    def __init__(self, pkgfile_path, path):
        self.is_archive = path.endswith('.a')
        self.is_shlib = self.so_regex.search(path)
        self.is_debug = path.endswith('.debug')

        # Do not follow debug info links
        output = subprocess.check_output('readelf --help', shell=True, encoding='utf8')
        flag = '--debug-dump=no-follow-links'
        extra_flags = [flag] if flag in output else []

        self.section_info = ElfSectionInfo(pkgfile_path, extra_flags)
        self.program_header_info = ElfProgramHeaderInfo(pkgfile_path, extra_flags)
        self.dynamic_section_info = ElfDynamicSectionInfo(pkgfile_path, extra_flags)
        self.symbol_table_info = ElfSymbolTableInfo(pkgfile_path, extra_flags)
        self.comment_section_info = ElfCommentInfo(pkgfile_path, extra_flags)

    def parsing_failed_reason(self):
        reasons = [self.section_info.parsing_failed_reason,
                   self.program_header_info.parsing_failed_reason,
                   self.dynamic_section_info.parsing_failed_reason,
                   self.symbol_table_info.parsing_failed_reason,
                   self.comment_section_info.parsing_failed_reason]
        reasons = [r for r in reasons if r]
        for reason in reasons:
            if self.NOT_ELF_ERROR in reason:
                return self.NOT_ELF_ERROR
        return '\n'.join(reasons) if reasons else None



#!/usr/bin/python3

import os

import requests

SUSE_EXCEPTIONS = """
AGPL-3.0 AGPL-3.0+ GFDL-1.1 GFDL-1.1+ GFDL-1.2 GFDL-1.2+ GFDL-1.3 GFDL-1.3+ GPL-3.0-with-GCC-exception \
GPL-2.0-with-classpath-exception GPL-2.0-with-font-exception SUSE-LGPL-2.1+-with-GCC-exception SUSE-NonFree \
GPL-1.0+ GPL-1.0 GPL-2.0+ GPL-2.0 GPL-3.0+ GPL-3.0 LGPL-2.0 LGPL-2.0+ LGPL-2.1+ LGPL-2.1 LGPL-3.0+ LGPL-3.0
"""
IGNORED_FOR_PLUS = (' with ', '-with-', ' or ')

licenses = requests.get('https://raw.githubusercontent.com/openSUSE/obs-service-format_spec_file/master/licenses_changes.txt')
licenses_exceptions = requests.get('https://raw.githubusercontent.com/openSUSE/obs-service-format_spec_file/master/licenses_exceptions.txt')

with open('../configs/openSUSE/licenses.toml', 'w') as wfile:
    script_name = os.path.basename(__file__)
    wfile.write('# Generated with %s script:\n' % script_name)
    wfile.write('ValidLicenses = [\n')
    added = set()
    for line in licenses.text.splitlines():
        if line == 'First line':
            continue
        name = line.strip().split('\t')[0]
        if name not in added:
            wfile.write(f'    "{name}",\n')
            added.add(name)

        if not name.endswith('+') and not any(i in name.lower() for i in IGNORED_FOR_PLUS):
            plus_name = f'{name}+'
            if plus_name not in added:
                wfile.write(f'    "{plus_name}",\n')
                added.add(plus_name)
    wfile.write('    # SUSE EXCEPTIONS\n')
    for name in SUSE_EXCEPTIONS.strip().split(' '):
        wfile.write(f'    "{name}",\n')
    wfile.write(']\n\n')

    # Licenses Exceptions
    wfile.write('ValidLicenseExceptions = [\n')
    for line in licenses_exceptions.text.splitlines():
        wfile.write(f'    "{line}",\n')
    wfile.write(']\n')



import re

from mockdata.mock_files import (
    DevelopmentPackage,
    DirectoryWithoutXPerm2Package,
    DirectoryWithoutXPermPackage,
    FileChecksPackage,
    FilesWithoutPermsPackage,
    FilesWithoutPermsTmpfilesPackage,
    FileZeroLengthPackage,
    IncorrectFSFAddress,
    MakefileJunkPackage,
    ManPagesPackage,
    ManualPagesPackage,
    NetmaskDebugsourcePackage,
    NonReadableGhostPackage,
    Python3PowerBrokenPackage,
    Python3PowerPackage,
    PythonShebangLinkOkPackage,
    PythonShebangLinkPackage,
    RustFilesPackage,
    Shlib1Package,
    Shlib2DevelPackage,
    SphinxInvPackage,
    TclPackage,
    TestDocumentationPackage,
    UnexpandedMacroFilesPackage,
)
import pytest
from rpmlint.checks.FilesCheck import FilesCheck
from rpmlint.checks.FilesCheck import pyc_magic_from_chunk, pyc_mtime_from_chunk
from rpmlint.checks.FilesCheck import python_bytecode_to_script as pbts
from rpmlint.checks.FilesCheck import script_interpreter as se
from rpmlint.filter import Filter

from Testing import CONFIG, get_tested_path


@pytest.fixture(scope='function', autouse=True)
def filescheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = FilesCheck(CONFIG, output)
    yield output, test


@pytest.fixture
def output(filescheck):
    output, _test = filescheck
    yield output


@pytest.fixture
def test(filescheck):
    _output, test = filescheck
    yield test


def test_pep3147():
    assert pbts('/usr/lib64/python3.4/__pycache__/__phello__.foo.cpython-34.pyc') == '/usr/lib64/python3.4/__phello__.foo.py'
    assert pbts('/usr/lib64/python3.4/__pycache__/__phello__.foo.cpython-34.pyo') == '/usr/lib64/python3.4/__phello__.foo.py'


def test_py2():
    assert pbts('/usr/lib/python2.7/site-packages/_pytest/main.pyc') == '/usr/lib/python2.7/site-packages/_pytest/main.py'
    assert pbts('/usr/lib/python2.7/site-packages/_pytest/main.pyo') == '/usr/lib/python2.7/site-packages/_pytest/main.py'


def test_pep0488():
    assert pbts('/usr/lib/python3.5/site-packages/__pycache__/pytest.cpython-35.opt-1.pyc') == '/usr/lib/python3.5/site-packages/pytest.py'
    assert pbts('/usr/lib/python3.5/site-packages/__pycache__/pytest.cpython-35.opt-2.pyc') == '/usr/lib/python3.5/site-packages/pytest.py'
    assert pbts('/usr/lib/python3.5/site-packages/__pycache__/pytest.cpython-35.pyc') == '/usr/lib/python3.5/site-packages/pytest.py'


def chunk_from_pyc(version, size=16):
    """Helper to get start of an example pyc file as bytes"""
    path = get_tested_path(f'pyc/__future__.cpython-{version}.pyc')
    with open(path, 'rb') as f:
        return f.read(size)


@pytest.mark.parametrize('package', [UnexpandedMacroFilesPackage])
def test_unexpanded_macros(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'unexpanded-macro' in out


@pytest.mark.parametrize('package,result', [
    (Python3PowerPackage, False),
    (Python3PowerBrokenPackage, True),
])
def test_python_bytecode_magic(package, result, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert ('python-bytecode-wrong-magic-value' in out) == result


@pytest.mark.parametrize('package', [TestDocumentationPackage])
def test_file_not_utf8_for_compression_algorithms(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'file-not-utf8 /usr/share/doc/packages/testdocumentation/README1.gz' in out
    assert 'file-not-utf8 /usr/share/doc/packages/testdocumentation/README2.bz2' in out
    assert 'file-not-utf8 /usr/share/doc/packages/testdocumentation/README3.xz' in out


@pytest.mark.parametrize('version, magic', ((36, 3379), (37, 3393)))
def test_pyc_magic_from_chunk(version, magic):
    chunk = chunk_from_pyc(version)
    assert pyc_magic_from_chunk(chunk) == magic


@pytest.mark.parametrize('version, mtime', ((36, 1513659236), (37, 1519778958)))
def test_pyc_mtime_from_chunk(version, mtime):
    chunk = chunk_from_pyc(version)
    assert pyc_mtime_from_chunk(chunk) == mtime


@pytest.mark.parametrize('package', [NetmaskDebugsourcePackage])
def test_devel_files(package, filescheck):
    output, test = filescheck
    test.check(package)
    assert len(output.results) == 5
    out = output.print_results(output.results)
    assert 'devel-file-in-non-devel-package' not in out
    assert 'incorrect-fsf-address' in out
    assert 'no-documentation' in out


@pytest.mark.parametrize('package', [IncorrectFSFAddress])
def test_incorrect_fsf_address(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'incorrect-fsf-address /usr/incorrect-fsf.txt' in out
    assert 'incorrect-fsf-address /usr/correct-fsf.txt' not in out


@pytest.mark.parametrize('package', [MakefileJunkPackage])
def test_makefile_junk(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'W: makefile-junk /usr/share/Makefile.am' in out
    assert out.count('W: makefile-junk') == 1


@pytest.mark.parametrize('package', [SphinxInvPackage])
def test_sphinx_inv_files(package, filescheck):
    output, test = filescheck
    test.check(package)
    assert not len(output.results)


@pytest.mark.parametrize('package', [FileChecksPackage])
def test_invalid_package(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'W: non-ghost-in-run /run/foo' in out
    assert 'W: systemd-unit-in-etc /etc/systemd/system/foo' in out
    assert 'W: udev-rule-in-etc /etc/udev/rules.d/foo' in out
    assert 'W: tmpfiles-conf-in-etc /etc/tmpfiles.d/foo' in out
    assert 'E: subdir-in-bin /bin/foo/bar' in out
    assert 'W: siteperl-in-perl-module /site_perl/foo' in out
    assert 'E: backup-file-in-package /~backup.rej' in out
    assert 'E: version-control-internal-file /.gitignore' in out
    assert 'E: htaccess-file /.htaccess' in out
    assert 'W: manifest-in-perl-module /usr/share/doc/perl-foo/MANIFEST' in out
    assert 'E: info-dir-file /usr/info/dir' in out


@pytest.mark.parametrize('package', [TclPackage])
def test_tcl_package(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: tcl-extension-file /usr/lib64/tcl/pkgIndex.tcl' in out


def test_script_interpreter():
    assert se(b'#!/bin/sh\n# Hello world!\n') == ('/bin/sh', '')
    assert se(b'#!/bin/bash -e\n') == ('/bin/bash', '-e')
    assert se(b'#! /usr/bin/perl -wT \n') == ('/usr/bin/perl', '-wT')
    assert se(b'#!/usr/bin/env python3 foo') == ('/usr/bin/env', 'python3 foo')
    assert se(b'# something here\n#!not a shebang') == (None, '')


def test_scm_regex():
    from rpmlint.checks.FilesCheck import scm_regex

    assert scm_regex.search('/foo/CVS/bar')
    assert scm_regex.search('/foo/RCS/bar')
    assert scm_regex.search('/bar/foo,v')
    assert scm_regex.search('bar/.svnignore')
    assert scm_regex.search('bar/.git/refs')


def test_lib_regex():
    from rpmlint.checks.FilesCheck import lib_regex

    # true matches
    assert all(
        lib_regex.search(x) for x in
        ('/lib/libnsl-2.26.so',
         '/usr/lib64/libgnomeui.so.3',
         '/lib64/libgcc_s.so.1'))

    # false positives
    assert not any(
        lib_regex.search(x) for x in
        ('/usr/share/gdb/auto-load/usr/lib/libglib-2.0.so.0.4600.1-gdb.py',
         '/usr/share/doc/findlib/lib-1.0.so',
         '/usr/lib64/libvulkan_radeon.so',
         '/usr/lib64/rsocket/binary',))


@pytest.mark.parametrize('package', [RustFilesPackage])
def test_rust_files(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: wrong-script-interpreter /etc/foo.rs' in out
    assert 'E: wrong-script-interpreter /etc/bar.rs' not in out


@pytest.mark.parametrize('package', [ManPagesPackage])
def test_manpages(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'manpage-not-compressed' in out
    assert 'no-manual-page-for-binary' not in out
    assert 'This manual page is not compressed with the bz2 compression' in out


@pytest.mark.parametrize('package', [DevelopmentPackage])
def test_provides_devel(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: non-devel-file-in-devel-package /usr/x.typelib' in out


@pytest.mark.parametrize('package,is_devel', [
    (Shlib1Package, False),
    (Shlib2DevelPackage, True),
])
def test_shlib(package, is_devel, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'library-without-ldconfig-postin' in out
    assert 'library-without-ldconfig-postun' in out
    if is_devel:
        assert ('non-devel-file-in-devel-package' in out)
    else:
        assert ('devel-file-in-non-devel-package' in out)


@pytest.mark.parametrize('package, files', [
    (FileZeroLengthPackage, [
        ('/usr/lib/emptyfile', True),
        ('/usr/lib/nonemptyfile', False),
        ('/etc/security/console.apps', False),
        ('/etc/security/console.apps/myapp', False),
        ('/usr/lib/.nosearch', False),
        ('/etc/share/doc/packages/file-zero-length/dummydoc', False),
        ('/usr/lib/python/__init__.py', False),
        ('/usr/lib/python/py.typed', False),
        ('/usr/lib/python/pypackagefromwheel-0.0.0.dist-info/REQUESTED', False),
        ('/usr/lib/ruby/gem.build_complete', False),
    ]),
])
def test_zero_length_ignore(package, files, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    for filename, show in files:
        assert filename in package.files
        assert (f'zero-length {filename}' in out) == show


@pytest.mark.parametrize('package', [ManualPagesPackage])
def test_manual_pages(package, filescheck):
    output, test = filescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: manual-page-in-subfolder /usr/share/man/man3/foo/bar/baz.3.gz' in out
    assert 'W: manpage-not-compressed bz2 /usr/share/man/man1/test.1.zst' in out
    assert 'E: bad-manual-page-folder /usr/share/man/man0p/foo.3.gz expected folder: man3' in out
    assert 'E: bad-manual-page-folder /usr/share/man/man3/some.3pm.gz' not in out


@pytest.mark.parametrize('package,is_ok', [
    (PythonShebangLinkPackage, False),
    (PythonShebangLinkOkPackage, True),
])
def test_shebang(package, is_ok, output, test):
    test.check(package)
    out = output.print_results(output.results)
    assert ('W: symlink-to-binary-with-shebang /usr/bin/testlink' not in out) == is_ok


@pytest.mark.parametrize('package, is_ok', [
    (DirectoryWithoutXPermPackage, False),
    (DirectoryWithoutXPerm2Package, False),
])
def test_directory_without_x_permission(package, is_ok, output, test):
    test.check(package)
    out = output.print_results(output.results)
    assert ('E: non-standard-dir-perm' not in out) == is_ok


@pytest.mark.parametrize('package', [FilesWithoutPermsPackage])
def test_files_without_perms(package, output, test):
    test.check(package)
    out = output.print_results(output.results)
    assert re.findall('W: zero-perms .*pipewire ', out)
    assert re.findall('W: zero-perms-ghost .*ghost_file', out)
    assert re.findall('W: zero-perms .*normal_file ', out)
    assert not re.findall('W: zero-perms .*normal_file_read ', out)
    assert not re.findall('W: zero-perms .*dir_read ', out)
    assert not re.findall('W: zero-perms-ghost .*ghost_file_read', out)


@pytest.mark.parametrize('package', [FilesWithoutPermsTmpfilesPackage])
def test_files_without_perms_tmpfiles(package, output, test):
    test.check(package)
    out = output.print_results(output.results)
    assert re.findall(r'W: zero-perms-ghost .*"%ghost %attr\(0644,root,root\) .*resolv.conf"', out)
    assert re.findall(r'W: zero-perms-ghost .*"%ghost %attr\(0755,root,group\) /run/netconfig"', out)
    assert not re.findall('W: zero-perms.*yp.conf ', out)


# https://github.com/rpm-software-management/rpmlint/issues/1287
@pytest.mark.parametrize('package', [NonReadableGhostPackage])
def test_non_readable_ghost_files(package, output, test):
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: non-readable /boohoo 0' not in out



from Testing import get_tested_mock_package


KRB5_CONF = """
d /var/lib/kerberos 		0755	root	root	-
d /var/lib/kerberos/krb5	0755	root	root	-
d /var/lib/kerberos/krb5/user	0755	root	root	-
d /var/lib/kerberos/krb5kdc	0755	root	root	-
C /var/lib/kerberos/krb5kdc/kdc.conf	0600 root root - /usr/share/kerberos/krb5kdc/kdc.conf
C /var/lib/kerberos/krb5kdc/kadm5.acl	0600 root root - /usr/share/kerberos/krb5kdc/kadm5.acl
C /var/lib/kerberos/krb5kdc/kadm5.dict	0600 root root - /usr/share/kerberos/krb5kdc/kadm5.dict
"""


SYSTEMD_TMPFILES_CONF = """
# create a directory with permissions 0770 owned by user foo and group bar
d /run/my_new_directory 0770 foo bar
"""


PREIN = """
[ -z "${TRANSACTIONAL_UPDATE}" -a -x /usr/bin/systemd-tmpfiles ] &&
    /usr/bin/systemd-tmpfiles --create /usr/lib/tmpfiles.d/systemd-tmpfiles.conf || :
"""

POSTIN = """
[ -z "${TRANSACTIONAL_UPDATE}" -a -x /usr/bin/systemd-tmpfiles ] &&
    /usr/bin/systemd-tmpfiles --create /usr/lib/tmpfiles.d/systemd-tmpfiles_correct.conf || :
"""


TempfiledPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/tmpfiles.d/krb5.conf': {'content': KRB5_CONF},
        '/usr/lib/tmpfiles.d/symlink.conf': {'linkto': '/usr/lib/tmpfiles.d/krb5.conf'}
    },
    header={},
)


SystemdTempfilesPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/tmpfiles.d/systemd-tmpfiles.conf': {'content': SYSTEMD_TMPFILES_CONF},
    },
    header={'PREIN': PREIN},
)


SystemdTempfilesOkPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/run/my_new_directory': {},
        '/usr/lib/tmpfiles.d/systemd-tmpfiles_correct.conf': {'content': SYSTEMD_TMPFILES_CONF},
    },
    header={'POSTIN': POSTIN},
)



from rpmlint.checks.AbstractCheck import AbstractCheck


class ConfigFilesCheck(AbstractCheck):
    """
    Check that configuration files are in a proper location and marked as
    'noreplace'.
    """
    def check_binary(self, pkg):
        for filename in pkg.config_files:
            self._check_non_confdir_files(pkg, filename)
            self._check_noreplace_files(pkg, filename)

    def _check_non_confdir_files(self, pkg, fn):
        """
        Check if the configuration file is in /etc or /var directory.

        Print a warning if it's not.
        """
        if not fn.startswith('/etc/') and not fn.startswith('/var/'):
            self.output.add_info('W', pkg,
                                 'non-etc-or-var-file-marked-as-conffile', fn)

    def _check_noreplace_files(self, pkg, fn):
        """
        Check if the configuration file has 'noreplace' flag.

        Print a warning if there is no 'noreplace' tag.
        """
        if fn not in pkg.noreplace_files:
            self.output.add_info('W', pkg,
                                 'conffile-without-noreplace-flag', fn)



from Testing import get_tested_mock_package

ErlangPackage = get_tested_mock_package(
    lazyload=True,
    files={
        '/usr/lib/erlang/m-no-CInf.beam': {
            'content-path': 'files/m-no-CInf.beam',
            'create_dirs': True,
        },
        '/usr/lib/erlang/m.beam': {
            'content-path': 'files/m.beam',
            'create_dirs': True,
        },
    },
    header={
        'requires': [
            'rpmlib(CompressedFileNames) <= 3.0.4-1',
            'rpmlib(FileDigests) <= 4.6.0-1',
            'rpmlib(PayloadFilesHavePrefix) <= 4.0-1',
            'rpmlib(PayloadIsZstd) <= 5.4.18-1',
        ],
    },
)



import subprocess
from xml.etree import ElementTree

from rpmlint.checks.AbstractCheck import AbstractFilesCheck
from rpmlint.helpers import ENGLISH_ENVIRONMENT


class AppDataCheck(AbstractFilesCheck):
    """
    check appdata files for format violations
    https://www.freedesktop.org/software/appstream/docs/
    """
    # default command, split here so we can mock it later
    cmd = 'appstream-util validate-relax --nonet '

    def __init__(self, config, output):
        super().__init__(config, output, r'/usr/share/appdata/.*\.(appdata|metainfo).xml$')

    def check_file(self, pkg, filename):
        root = pkg.dir_name()
        f = root + filename
        cmd = self.cmd + f

        validation_failed = False
        try:
            r = subprocess.run(cmd.split(), env=ENGLISH_ENVIRONMENT,
                               stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
            if r.returncode != 0:
                validation_failed = True
        except FileNotFoundError:
            # appstream-util is not installed
            # validate the xml format only
            try:
                ElementTree.parse(f)
            except ElementTree.ParseError:
                validation_failed = True
        if validation_failed:
            self.output.add_info('E', pkg, 'invalid-appdata-file', filename)



import stat
import subprocess

from rpmlint.checks.AbstractCheck import AbstractFilesCheck
from rpmlint.helpers import ENGLISH_ENVIRONMENT


class BashismsCheck(AbstractFilesCheck):
    def __init__(self, config, output):
        super().__init__(config, output, r'.*')
        self.use_threads = True
        self._detect_early_fail_option()
        self.file_cache = {}

    def reset(self):
        super().reset()
        self.file_cache = {}

    def _detect_early_fail_option(self):
        output = subprocess.check_output('checkbashisms --help',
                                         shell=True, encoding='utf8')
        # FIXME: remove in the future
        self.use_early_fail = '[-e]' in output

    def check_file(self, pkg, filename):
        root = pkg.dir_name()
        pkgfile = pkg.files[filename]
        filepath = root + filename

        # We only care about the real files that state they are shell scripts
        if not (stat.S_ISREG(pkgfile.mode) and
                pkgfile.magic.startswith('POSIX shell script')):
            return

        # There are package likes Linux kernel where there are common
        # shell scripts present in multiple packages
        # (kernel-source, kernel-source-vanilla).
        if pkgfile.md5 not in self.file_cache:
            self.file_cache[pkgfile.md5] = list(self.check_bashisms(pkg, filepath, filename))

        for warning in self.file_cache[pkgfile.md5]:
            self.output.add_info('W', pkg, warning, filename)

    def check_bashisms(self, pkg, filepath, filename):
        """
        Run dash and then checkbashism on file

        We need to see if it is valid syntax of bash and if there are no
        potential bash issues.
        Return a warning message or None if there is no problem.
        """
        try:
            r = subprocess.run(['dash', '-n', filepath],
                               stderr=subprocess.DEVNULL,
                               env=ENGLISH_ENVIRONMENT)
            if r.returncode == 2:
                yield 'bin-sh-syntax-error'
            elif r.returncode == 127:
                raise FileNotFoundError(filename)
        except UnicodeDecodeError:
            pass

        try:
            cmd = ['checkbashisms', filepath]
            # --early-fail option can rapidly speed up the check
            if self.use_early_fail:
                cmd.append('-e')
            r = subprocess.run(cmd,
                               stderr=subprocess.DEVNULL,
                               env=ENGLISH_ENVIRONMENT)
            if r.returncode == 1:
                yield 'potential-bashisms'
            elif r.returncode == 2:
                raise FileNotFoundError(filename)
        except UnicodeDecodeError:
            pass



from pathlib import Path
import re
import subprocess
from urllib.parse import urlparse

import rpm
from rpmlint import pkg as Pkg
from rpmlint.checks.AbstractCheck import AbstractCheck
from rpmlint.helpers import ENGLISH_ENVIRONMENT, readlines

# Don't check for hardcoded library paths in biarch packages
DEFAULT_BIARCH_PACKAGES = '^(gcc|glibc)'


def re_tag_compile(tag):
    rpm_tag = fr'^{tag}\s*:\s*(\S.*?)\s*$'
    return re.compile(rpm_tag, re.IGNORECASE)


patch_regex = re_tag_compile(r'Patch(\d*)')
# rpm 4.20 doesn't support %patchN anymore, we should warn about this
applied_patch_rpm420_regex = re.compile(r'^%patch(\d+)')
applied_patch_regex = re.compile(r'^%patch\s*(\d*)')
applied_patch_p_regex = re.compile(r'\s-P\s*(\d+)\b')
applied_patch_pipe_regex = re.compile(r'\s%\{PATCH(\d+)\}\s*\|\s*(%\{?__)?patch\b')
applied_patch_i_regex = re.compile(r'(?:%\{?__)?patch\}?.*?\s+(?:<|-i)\s+%\{PATCH(\d+)\}')
source_dir_regex = re.compile(r'^[^#]*(\$RPM_SOURCE_DIR|%{?_sourcedir}?)')
obsolete_tags_regex = re_tag_compile(r'(?:Serial|Copyright)')
buildroot_regex = re_tag_compile('BuildRoot')
prefix_regex = re_tag_compile('Prefix')
packager_regex = re_tag_compile('Packager')
buildarch_regex = re_tag_compile('BuildArch(?:itectures)?')
buildprereq_regex = re_tag_compile('BuildPreReq')
prereq_regex = re_tag_compile(r'PreReq(\(.*\))')

make_check_regex = re.compile(r'(^|\s|%{?__)make}?\s+(check|test)')
rm_regex = re.compile(r'(^|\s)((.*/)?rm|%{?__rm}?) ')
rpm_buildroot_regex = re.compile(r'^[^#]*?(?:(\\*)\${?RPM_BUILD_ROOT}?|(%+){?buildroot}?)')
configure_libdir_spec_regex = re.compile(r'ln |\./configure[^#]*--libdir=(\S+)[^#]*')
lib_package_regex = re.compile(r'^%package.*\Wlib')
ifarch_regex = re.compile(r'^\s*%ifn?arch\s')
if_regex = re.compile(r'^\s*%if\s')
endif_regex = re.compile(r'^\s*%endif\b')
biarch_package_regex = re.compile(DEFAULT_BIARCH_PACKAGES)
libdir_regex = re.compile(r'%{?_lib(?:dir)?\}?\b')
section_regexs = {x: re.compile('^%' + x + r'(?:\s|$)')
                  for x in ('build', 'changelog', 'check', 'clean', 'description', 'files',
                            'install', 'package', 'prep') + Pkg.RPM_SCRIPTLETS}
deprecated_grep_regex = re.compile(r'\b[ef]grep\b')

# Only check for /lib, /usr/lib, /usr/X11R6/lib
# TODO: better handling of X libraries and modules.
hardcoded_library_paths = '(/lib|/usr/lib|/usr/X11R6/lib/(?!([^/]+/)+)[^/]*\\.([oa]|la|so[0-9.]*))'
hardcoded_library_path_regex = re.compile(r'^[^#]*((^|\s+|\.\./\.\.|\${?RPM_BUILD_ROOT}?|%{?buildroot}?|%{?_prefix}?)' + hardcoded_library_paths + r'(?=[\s;/])([^\s,;]*))')

DEFINE_RE = r'(^|\s)%(define|global)\s+'
depscript_override_regex = re.compile(DEFINE_RE + r'__find_(requires|provides)\s')
depgen_disable_regex = re.compile(DEFINE_RE + r'_use_internal_dependency_generator\s+0')
patch_fuzz_override_regex = re.compile(DEFINE_RE + r'_default_patch_fuzz\s+(\d+)')

# See https://bugzilla.redhat.com/488146 for details
indent_spaces_regex = re.compile('( \t|(^|\t)([^\t]{8})*[^\t]{4}[^\t]?([^\t][^\t.!?]|[^\t]?[.!?] )  )')

requires_regex = re.compile(r'^(?:Build)?(?:Pre)?Req(?:uires)?(?:\([^\)]+\))?:\s*(.*)', re.IGNORECASE)
provides_regex = re.compile(r'^Provides(?:\([^\)]+\))?:\s*(.*)', re.IGNORECASE)
obsoletes_regex = re.compile(r'^Obsoletes:\s*(.*)', re.IGNORECASE)
conflicts_regex = re.compile(r'^(?:Build)?Conflicts:\s*(.*)', re.IGNORECASE)

declarative_regex = re.compile(r'^BuildSystem:\s*(.*)', re.IGNORECASE)

compop_regex = re.compile(r'[<>=]')

setup_regex = re.compile(r'%setup\b')  # intentionally no whitespace before!
setup_q_regex = re.compile(r' -[A-Za-z]*q')
setup_t_regex = re.compile(r' -[A-Za-z]*T')
setup_ab_regex = re.compile(r' -[A-Za-z]*[ab]')
autosetup_regex = re.compile(r'^\s*%autosetup(\s.*|$)')
autosetup_n_regex = re.compile(r' -[A-Za-z]*N')
autopatch_regex = re.compile(r'^\s*%autopatch(?:\s|$)')

filelist_regex = re.compile(r'\s+-f\s+\S+')
pkgname_regex = re.compile(r'\s+(?:-n\s+)?(\S+)')
tarball_regex = re.compile(r'\.(?:t(?:ar|[glx]z|bz2?)|zip)\b', re.IGNORECASE)

python_setup_test_regex = re.compile(r'^[^#]*(setup.py test)')
python_setup_install_regex = re.compile(r'^[^#]*(setup.py install|%\{?py(thon)?\d*_install)')
python_module_def_regex = re.compile(r'^[^#]*%{\?!python_module:%define python_module()')
python_sitelib_glob_regex = re.compile(r'^[^#]*%{python_site(lib|arch)}/\*\s*$')

shared_dir_glob_regex = re.compile(r'^[^#]*%{_(?:bin|data|doc|include|man)dir}/\*\s*$')

# %suse_update_desktop_file deprecation
# https://lists.opensuse.org/archives/list/packaging@lists.opensuse.org/message/TF4QO7ECOSEDHBFI5YDEA3OF4RNSI7D7/
suse_update_desktop_file_regex = re.compile(r'^BuildRequires:\s*update-desktop-files', re.IGNORECASE)

UNICODE_NBSP = '\xa0'


def unversioned(deps):
    """Yield unversioned dependency names from the given list."""
    for dep in deps:
        if not dep[1]:
            yield dep[0]


def contains_buildroot(line):
    """Check if the given line contains use of rpm buildroot."""
    res = rpm_buildroot_regex.search(line)
    if res and \
       (not res.group(1) or len(res.group(1)) % 2 == 0) and \
       (not res.group(2) or len(res.group(2)) % 2 != 0):
        return True
    return False


class SpecCheck(AbstractCheck):
    """Contain check methods that catch errors and warnings in a specfile."""

    def __init__(self, config, output):
        super().__init__(config, output)
        self._spec_file = None
        self._spec_name = None
        self.valid_groups = config.configuration['ValidGroups']
        self.output.error_details.update({'non-standard-group':
                                         """The value of the Group tag in the package is not valid.  Valid groups are:
                                         '%s'.""" % ', '.join(self.valid_groups)})
        self.hardcoded_lib_path_exceptions_regex = re.compile(config.configuration['HardcodedLibPathExceptions'])

        self._default_state()

    def _default_state(self):
        # Default state
        self.patches = {}
        self.applied_patches = []
        self.applied_patches_ifarch = []
        self.patches_auto_applied = False
        self.source_dir = False
        self.buildroot = False
        self.configure_linenum = None
        self.configure_cmdline = ''
        self.mklibname = False
        self.is_lib_pkg = False
        self.if_depth = 0
        self.ifarch_depth = -1
        self.depscript_override = False
        self.depgen_disabled = False
        self.patch_fuzz_override = False
        self.indent_spaces = 0
        self.indent_tabs = 0
        self.section = {}
        self.declarative = False

        self.current_section = 'package'
        # None == main package
        self.current_package = None
        self.package_noarch = {}

    def reset(self):
        self._spec_file = None
        self._spec_name = None
        self._default_state()

    def check_source(self, pkg):
        """Find specfile in SRPM and run spec file related checks."""
        wrong_spec = False
        self._spec_file = None
        self._spec_name = None

        # Check if a specfile exist in a specified path
        for fname, pkgfile in pkg.files.items():
            if fname.endswith('.spec'):
                self._spec_file = pkgfile.path
                self._spec_name = pkgfile.name
                if fname == pkg.name + '.spec':
                    wrong_spec = False
                    break
                else:
                    wrong_spec = True

        # method call
        self._check_no_spec_file(pkg)
        self._check_invalid_spec_name(pkg, wrong_spec)

        if self._spec_file:
            # check content of spec file
            with Pkg.FakePkg(self._spec_file) as package:
                self.check_spec(package)

    def check_spec(self, pkg):
        """Find specfile in specified path and run spec file related checks."""
        self._spec_file = pkg.name
        self._spec_file_dir = str(Path(self._spec_file).parent)

        # method call
        self._check_non_utf8_spec_file(pkg)

        self.pkg = pkg
        self.spec_only = isinstance(pkg, Pkg.FakePkg)

        spec_lines = readlines(self._spec_file)
        # Analyse specfile line by line to check for (E)rrors or (W)arnings
        # And initialize the SpecCheck instance for following checks
        self._check_lines(spec_lines)

        # Run checks for whole package
        self._check_no_buildroot_tag(pkg, self.buildroot)

        if not self.declarative:
            self._check_no_s_section(pkg, self.section)

        self._check_superfluous_clean_section(pkg, self.section)
        self._check_more_than_one_changelog_section(pkg, self.section)
        self._check_lib_package_without_mklibname(pkg, self.is_lib_pkg, self.mklibname)
        self._check_descript_without_disabling_depgen(pkg, self.depscript_override,
                                                      self.depgen_disabled)
        self._check_patch_fuzz_is_changed(pkg, self.patch_fuzz_override)
        self._check_mixed_use_of_space_and_tabs(pkg, self.indent_spaces, self.indent_tabs)
        self.check_ifarch_and_not_applied_patches(pkg, self.patches_auto_applied, self.patches,
                                                  self.applied_patches_ifarch, self.applied_patches)

        # Checks below require a real spec file
        if not self._spec_file:
            return
        self._check_specfile_error(pkg)
        self._check_invalid_url(pkg, rpm)

    def _check_no_spec_file(self, pkg):
        """Check if no spec file is found in RPM meta data."""
        if not self._spec_file:
            self.output.add_info('E', pkg, 'no-spec-file')

    def _check_invalid_spec_name(self, pkg, wrong_spec):
        """Check if spec file has same name as the 'Name: ' tag."""
        if wrong_spec and self._spec_file:
            self.output.add_info('E', pkg, 'invalid-spec-name')

    def _check_non_utf8_spec_file(self, pkg):
        """Check if spec file has UTF-8 character encoding."""
        if self._spec_file and not Pkg.is_utf8(self._spec_file):
            self.output.add_info('E', pkg, 'non-utf8-spec-file',
                                 self._spec_name or self._spec_file)

    def _check_no_buildroot_tag(self, pkg, buildroot):
        """Check if BuildRoot tag is used in the specfile."""
        if not buildroot:
            self.output.add_info('W', pkg, 'no-buildroot-tag')

    def _check_no_s_section(self, pkg, section):
        """Check if there is no (%prep, %build, %install, %check)
        in the specfile.
        """
        for sec in ('prep', 'build', 'install', 'check'):
            if not section.get(sec):
                self.output.add_info('W', pkg, 'no-%%%s-section' % sec)

    def _check_superfluous_clean_section(self, pkg, section):
        """Check for a superfluous %clean section in the specfile.
        """
        if section.get('clean'):
            self.output.add_info('E', pkg, 'superfluous-%clean-section')

    def _check_more_than_one_changelog_section(self, pkg, section):
        """Check if specfile has more than one %changelog.
        prep, build, install, check prevented by rpmbuild 4.4
        """
        if section.get('changelog', 0) > 1:
            self.output.add_info('W', pkg, 'more-than-one-%changelog-section')

    def _check_lib_package_without_mklibname(self, pkg, is_lib_pkg, mklibname):
        """Check if package name is built using %mklibname to allow lib64 and lib32
        coexistence. This check is specific to Mandriva and it's derivatives,
        check issue #9 in rpm-software-management/rpmlint/issues
        """
        if is_lib_pkg and not mklibname:
            self.output.add_info('E', pkg, 'lib-package-without-%mklibname')

    def _check_descript_without_disabling_depgen(self, pkg, depscript_override, depgen_disabled):
        """Check if specfile has %define _use_internal_dependency_generator set to 0
        to disable it, or does not have define __find_provides/requires.
        """
        if depscript_override and not depgen_disabled:
            self.output.add_info('W', pkg, 'depscript-without-disabling-depgen')

    def _check_patch_fuzz_is_changed(self, pkg, patch_fuzz_override):
        """Check if specfile has internal patch fuzz was changed."""
        if patch_fuzz_override:
            self.output.add_info('W', pkg, 'patch-fuzz-is-changed')

    def _check_mixed_use_of_space_and_tabs(self, pkg, indent_spaces, indent_tabs):
        """Check if specfile has mixed uses of spaces and tabs."""
        if indent_spaces and indent_tabs:
            pkg.current_linenum = max(indent_spaces, indent_tabs)
            self.output.add_info('W', pkg, 'mixed-use-of-spaces-and-tabs',
                                 '(spaces: line %d, tab: line %d)' %
                                 (indent_spaces, indent_tabs))
            pkg.current_linenum = None

    def check_ifarch_and_not_applied_patches(self, pkg, patches_auto_applied,
                                             patches, applied_patches_ifarch, applied_patches):
        """Check if specfile has a patch applied inside an %ifarch block.
        and check if a patch was included but not applied."""
        if not patches_auto_applied:
            for pnum, pfile in patches.items():
                if pnum in applied_patches_ifarch:
                    self.output.add_info('W', pkg, '%ifarch-applied-patch',
                                         'Patch%d:' % pnum, pfile)

                # Check if a patch is included in specfile but was not applied.
                if pnum not in applied_patches:
                    self.output.add_info('W', pkg, 'patch-not-applied',
                                         'Patch%d:' % pnum, pfile)

    def _check_specfile_error(self, pkg):
        """It parse the specfile with rpm and forward errors to rpmlint output."""

        # We'd like to parse the specfile only once using python bindings,
        # but it seems errors from rpmlib get logged to stderr and we can't
        # capture and print them nicely, so we do it once each way :P
        try:
            outcmd = subprocess.run(
                ('rpm', '-q', '--qf=', '-D', '_sourcedir %s' % self._spec_file_dir,
                 '--specfile', self._spec_file), stderr=subprocess.PIPE, encoding='utf8', env=ENGLISH_ENVIRONMENT)

            for line in outcmd.stderr.splitlines():
                line = line.strip()
                if not line:
                    continue
                if line and 'warning:' not in line:
                    self.output.add_info('E', pkg, 'specfile-error', line)
                else:
                    self.output.add_info('W', pkg, 'specfile-warning', line)
        except UnicodeDecodeError as e:
            self.output.add_info('E', pkg, 'specfile-error', str(e))

    def _check_invalid_url(self, pkg, rpm):
        """Check if specfile has an invalid url."""
        # grab sources and patches from parsed spec object to get
        # them with macros expanded for URL checking
        spec_obj = None
        rpm.addMacro('_sourcedir', self._spec_file_dir)
        try:
            transaction_set = rpm.TransactionSet()
            spec_obj = transaction_set.parseSpec(str(self._spec_file))
        except (ValueError, rpm.error) as e:
            self.output.add_info('E', pkg, 'specfile-error', str(e).strip(),
                                 str(self._spec_file))
        rpm.delMacro('_sourcedir')
        if spec_obj:
            for src in spec_obj.sources:
                (url, num, flags) = src
                (scheme, netloc) = urlparse(url)[0:2]
                if flags & 1:  # rpmspec.h, rpm.org ticket #123
                    srctype = 'Source'
                else:
                    srctype = 'Patch'
                tag = f'{srctype}{num}'
                if scheme and netloc:
                    continue
                elif srctype == 'Source' and tarball_regex.search(url):
                    self.output.add_info('W', pkg, 'invalid-url', '%s:' % tag, url)

    def _check_lines(self, lines):
        # gather info from spec lines
        self.pkg.current_linenum = 0
        for line in lines:
            self.pkg.current_linenum += 1
            self._check_line(line)
        # Last line read is not useful after this point
        self.pkg.current_linenum = None

    def _check_line(self, line):
        """
        Run check methods for this line.
        """

        self._checkline_declarative(line)
        self._checkline_break_space(line)
        if self._checkline_section(line):
            return
        self._checkline_buildroot_usage(line)
        self._checkline_make_check(line)
        self._checkline_setup(line)
        self._checkline_autopatch(line)
        self._checkline_applied_patch(line)
        self._checkline_sourcedir(line)
        self._checkline_configure(line)
        self._checkline_hardcoded_library_path(line)
        self._checkline_mklibname(line)
        self._checkline_package(line)
        self._checkline_changelog(line)
        self._checkline_files(line)
        self._checkline_indent(line)
        self._checkline_deprecated_grep(line)
        self._checkline_valid_groups(line)
        self._checkline_macros_in_comments(line)
        self._checkline_python_setup_test(line)
        self._checkline_python_setup_install(line)
        self._checkline_python_module_def(line)
        self._checkline_python_sitelib_glob(line)
        self._checkline_shared_dir_glob(line)

        # If statement, starts
        if ifarch_regex.search(line):
            self.if_depth = self.if_depth + 1
            self.ifarch_depth = self.if_depth
        elif if_regex.search(line):
            self.if_depth = self.if_depth + 1

        # If statement, ends
        elif endif_regex.search(line):
            if self.ifarch_depth == self.if_depth:
                self.ifarch_depth = -1
            self.if_depth = self.if_depth - 1

    # line checks methods

    def _checkline_declarative(self, line):
        # Do not override if we found the regex in previous lines
        if self.declarative:
            return
        self.declarative = bool(declarative_regex.search(line))
        if self.declarative:
            # we assume implicit %prep
            self.patches_auto_applied = True

    def _checkline_break_space(self, line):
        char = line.find(UNICODE_NBSP)
        if char != -1:
            self.output.add_info('W', self.pkg, 'non-break-space', 'line %s, char %d' %
                                 (self.pkg.current_linenum, char))

    def _checkline_section(self, line):
        section_marker = False
        for sec, regex in section_regexs.items():
            res = regex.search(line)
            if res:
                self.current_section = sec
                section_marker = True
                self.section[sec] = self.section.get(sec, 0) + 1
                if sec in ('package', 'files'):
                    rest = filelist_regex.sub('', line[res.end() - 1:])
                    res = pkgname_regex.search(rest)
                    if res:
                        self.current_package = res.group(1)
                    else:
                        self.current_package = None
                break

        if section_marker:
            if not self.is_lib_pkg and lib_package_regex.search(line):
                self.is_lib_pkg = True
            return True

    def _checkline_buildroot_usage(self, line):
        if (self.current_section in Pkg.RPM_SCRIPTLETS + ('prep', 'build') and
                contains_buildroot(line)):
            self.output.add_info('E', self.pkg, 'rpm-buildroot-usage', '%' + self.current_section,
                                 line[:-1].strip())

    def _checkline_make_check(self, line):
        if make_check_regex.search(line) and self.current_section not in \
                ('check', 'changelog', 'package', 'description'):
            self.output.add_info('W', self.pkg, 'make-check-outside-check-section',
                                 line[:-1])

    def _checkline_setup(self, line):
        # %setup check
        if setup_regex.match(line):
            if not setup_q_regex.search(line):
                # Don't warn if there's a -T without -a or -b
                if setup_t_regex.search(line):
                    if setup_ab_regex.search(line):
                        self.output.add_info('W', self.pkg, 'setup-not-quiet')
                else:
                    self.output.add_info('W', self.pkg, 'setup-not-quiet')

            if self.current_section != 'prep':
                self.output.add_info('W', self.pkg, 'setup-not-in-prep')
            return

        res = autosetup_regex.search(line)
        if res:
            if not autosetup_n_regex.search(res.group(1)):
                self.patches_auto_applied = True
            if self.current_section != 'prep':
                self.output.add_info('W', self.pkg, '%autosetup-not-in-prep')

    def _checkline_autopatch(self, line):
        # %autopach check
        if autopatch_regex.search(line):
            self.patches_auto_applied = True
            if self.current_section != 'prep':
                self.output.add_info('W', self.pkg, '%autopatch-not-in-prep')

    def _checkline_applied_patch(self, line):
        # Check for %patch -P
        res = applied_patch_regex.search(line)
        if res:
            # Check for %patchN (not supported by rpm >= 4.20)
            if applied_patch_rpm420_regex.match(line):
                self.output.add_info('E', self.pkg, 'patch-macro-old-format')

            pnum = res.group(1) or 0
            for tmp in applied_patch_p_regex.findall(line) or [pnum]:
                pnum = int(tmp)
                self.applied_patches.append(pnum)
                if self.ifarch_depth > 0:
                    self.applied_patches_ifarch.append(pnum)
            return

        # Check for %{PATCH0} | patch
        res = applied_patch_pipe_regex.search(line)
        if res:
            pnum = int(res.group(1))
            self.applied_patches.append(pnum)
            if self.ifarch_depth > 0:
                self.applied_patches_ifarch.append(pnum)
            return

        # Check for patch < %{PATCH0}
        res = applied_patch_i_regex.search(line)
        if res:
            pnum = int(res.group(1))
            self.applied_patches.append(pnum)
            if self.ifarch_depth > 0:
                self.applied_patches_ifarch.append(pnum)
            return

    def _checkline_sourcedir(self, line):
        if self.source_dir:
            return

        res = source_dir_regex.search(line)
        if res:
            self.source_dir = True
            self.output.add_info('E', self.pkg, 'use-of-RPM_SOURCE_DIR')

    def _checkline_configure(self, line):
        if self.configure_linenum:
            if self.configure_cmdline[-1] == '\\':
                self.configure_cmdline = self.configure_cmdline[:-1] + line.strip()
            else:
                res = configure_libdir_spec_regex.search(self.configure_cmdline)
                if not res:
                    # Hack to get the correct (start of ./configure) line
                    # number displayed:
                    real_linenum = self.pkg.current_linenum
                    self.pkg.current_linenum = self.configure_linenum
                    self.output.add_info('W', self.pkg, 'configure-without-libdir-spec')
                    self.pkg.current_linenum = real_linenum
                elif res.group(1):
                    res = re.match(hardcoded_library_paths, res.group(1))
                    if res:
                        self.output.add_info('E', self.pkg, 'hardcoded-library-path',
                                             res.group(1), 'in configure options')
                self.configure_linenum = None

        hash_pos = line.find('#')

        if self.current_section != 'changelog':
            cfg_pos = line.find('./configure')
            if cfg_pos != -1 and (hash_pos == -1 or hash_pos > cfg_pos):
                # store line where it started
                self.configure_linenum = self.pkg.current_linenum
                self.configure_cmdline = line.strip()

    def _checkline_hardcoded_library_path(self, line):
        res = hardcoded_library_path_regex.search(line)
        if self.current_section != 'changelog' and res and not \
                (biarch_package_regex.match(self.pkg.name) or
                 self.hardcoded_lib_path_exceptions_regex.search(
                     res.group(1).lstrip())):
            self.output.add_info('E', self.pkg, 'hardcoded-library-path', 'in',
                                 res.group(1).lstrip())

    def _checkline_mklibname(self, line):
        self.mklibname = '%mklibname' in line

    # line checks package methods
    def _checkline_package_patch(self, line):
        # Would be cleaner to get sources and patches from the
        # specfile parsed in Python (see below), but we want to
        # catch %ifarch'd etc ones as well, and also catch these when
        # the specfile is not parseable.
        res = patch_regex.search(line)
        if res:
            pnum = int(res.group(1) or 0)
            self.patches[pnum] = res.group(2)

    def _checkline_package_obsolete_tags(self, line):
        res = obsolete_tags_regex.search(line)
        if res:
            self.output.add_info('W', self.pkg, 'obsolete-tag', res.group(1))

    def _checkline_package_buildroot(self, line):
        res = buildroot_regex.search(line)
        if res:
            self.buildroot = True
            if res.group(1).startswith('/'):
                self.output.add_info('W', self.pkg, 'hardcoded-path-in-buildroot-tag',
                                     res.group(1))

    def _checkline_package_buildarch(self, line):
        res = buildarch_regex.search(line)
        if res:
            if res.group(1) != 'noarch':
                self.output.add_info('E', self.pkg,
                                     'buildarch-instead-of-exclusivearch-tag',
                                     res.group(1))
            else:
                self.package_noarch[self.current_package] = True

    def _checkline_package_packager(self, line):
        res = packager_regex.search(line)
        if res:
            self.output.add_info('W', self.pkg, 'hardcoded-packager-tag', res.group(1))

    def _checkline_package_prefix(self, line):
        res = prefix_regex.search(line)
        if res and not res.group(1).startswith('%'):
            self.output.add_info('W', self.pkg, 'hardcoded-prefix-tag', res.group(1))

    def _checkline_package_prereq(self, line):
        res = prereq_regex.search(line)
        if res:
            self.output.add_info('E', self.pkg, 'prereq-use', res.group(2))

    def _checkline_package_buildprereq(self, line):
        res = buildprereq_regex.search(line)
        if res:
            self.output.add_info('E', self.pkg, 'buildprereq-use', res.group(1))

    def _checkline_package_requires(self, line):
        res = requires_regex.search(line)
        if res:
            reqs = Pkg.parse_deps(res.group(1))
            deptoken = Pkg.has_forbidden_controlchars(reqs)
            if deptoken:
                self.output.add_info('E', self.pkg,
                                     'forbidden-controlchar-found',
                                     f'Requires: {deptoken}')
            for req in unversioned(reqs):
                if compop_regex.search(req):
                    self.output.add_info('W', self.pkg,
                                         'comparison-operator-in-deptoken',
                                         req)

    def _checkline_package_provides(self, line):
        res = provides_regex.search(line)
        if res:
            provs = Pkg.parse_deps(res.group(1))
            deptoken = Pkg.has_forbidden_controlchars(provs)
            if deptoken:
                self.output.add_info('E', self.pkg,
                                     'forbidden-controlchar-found',
                                     f'Provides: {deptoken}')
            for prov in unversioned(provs):
                if not prov.startswith('/'):
                    self.output.add_info('W', self.pkg, 'unversioned-explicit-provides',
                                         prov)
                if compop_regex.search(prov):
                    self.output.add_info('W', self.pkg,
                                         'comparison-operator-in-deptoken',
                                         prov)

    def _checkline_package_obsoletes(self, line):
        res = obsoletes_regex.search(line)
        if res:
            obses = Pkg.parse_deps(res.group(1))
            deptoken = Pkg.has_forbidden_controlchars(obses)
            if deptoken:
                self.output.add_info('E', self.pkg,
                                     'forbidden-controlchar-found',
                                     f'Obsoletes: {deptoken}')
            for obs in unversioned(obses):
                if not obs.startswith('/'):
                    self.output.add_info('W', self.pkg, 'unversioned-explicit-obsoletes',
                                         obs)
                if compop_regex.search(obs):
                    self.output.add_info('W', self.pkg,
                                         'comparison-operator-in-deptoken',
                                         obs)

    def _checkline_package_conflicts(self, line):
        res = conflicts_regex.search(line)
        if res:
            confs = Pkg.parse_deps(res.group(1))
            deptoken = Pkg.has_forbidden_controlchars(confs)
            if deptoken:
                self.output.add_info('E', self.pkg,
                                     'forbidden-controlchar-found',
                                     f'Conflicts: {deptoken}')
            for conf in unversioned(confs):
                if compop_regex.search(conf):
                    self.output.add_info('W', self.pkg,
                                         'comparison-operator-in-deptoken',
                                         conf)

    def _checkline_package(self, line):
        if self.current_section != 'package':
            return

        self._checkline_package_patch(line)
        self._checkline_package_obsolete_tags(line)
        self._checkline_package_buildroot(line)
        self._checkline_package_buildarch(line)
        self._checkline_package_packager(line)
        self._checkline_package_prefix(line)
        self._checkline_package_prereq(line)
        self._checkline_package_buildprereq(line)
        self._checkline_package_requires(line)
        self._checkline_package_provides(line)
        self._checkline_package_obsoletes(line)
        self._checkline_package_conflicts(line)

        self._checkline_forbidden_controlchars(line)
        self._check_suse_update_desktop_file(line)

    def _checkline_changelog(self, line):
        if self.current_section == 'changelog':
            deptoken = Pkg.has_forbidden_controlchars(line)
            if deptoken:
                self.output.add_info('E', self.pkg,
                                     'forbidden-controlchar-found',
                                     '%%changelog: %s' % deptoken)
            for match in self.macro_regex.findall(line):
                res = re.match('%+', match)
                if len(res.group(0)) % 2 and match != '%autochangelog' and match != '%{autochangelog}':
                    self.output.add_info('W', self.pkg, 'macro-in-%changelog', match)
        else:
            if not self.depscript_override:
                self.depscript_override = \
                    depscript_override_regex.search(line) is not None
            if not self.depgen_disabled:
                self.depgen_disabled = \
                    depgen_disable_regex.search(line) is not None
            if not self.patch_fuzz_override:
                self.patch_fuzz_override = \
                    patch_fuzz_override_regex.search(line) is not None

    def _checkline_files(self, line):
        # TODO: check scriptlets for these too?
        if (self.current_section == 'files' and
                (self.package_noarch.get(self.current_package) or
                    (self.current_package not in self.package_noarch and self.package_noarch.get(None)))):
            res = libdir_regex.search(line)
            if res:
                pkgname = self.current_package
                if pkgname is None:
                    pkgname = '(main package)'
                self.output.add_info('W', self.pkg, 'libdir-macro-in-noarch-package',
                                     pkgname, line.rstrip())

    def _checkline_indent(self, line):
        if not self.indent_tabs and '\t' in line:
            self.indent_tabs = self.pkg.current_linenum
        if not self.indent_spaces and indent_spaces_regex.search(line):
            self.indent_spaces = self.pkg.current_linenum

    def _checkline_deprecated_grep(self, line):
        # Check if egrep or fgrep is used
        if self.current_section not in \
                ('package', 'changelog', 'description', 'files'):
            greps = deprecated_grep_regex.findall(line)
            if greps:
                self.output.add_info('W', self.pkg, 'deprecated-grep', greps)

    def _checkline_valid_groups(self, line):
        # If not checking spec file only, we're checking one inside a
        # SRPM -> skip this check to avoid duplicate warnings (#167)

        if self.spec_only and self.valid_groups and \
           line.lower().startswith('group:'):
            group = line[6:].strip()
            if group not in self.valid_groups:
                self.output.add_info('W', self.pkg, 'non-standard-group', group)

    def _checkline_macros_in_comments(self, line):
        hash_pos = line.find('#')
        # Test if there are macros in comments
        if hash_pos != -1 and \
                (hash_pos == 0 or line[hash_pos - 1] in (' ', '\t')):

            comment = line[hash_pos + 1:]
            # Ignore special comments like #!BuildIgnore
            if comment and comment[0] == '!':
                return

            for match in self.macro_regex.findall(comment):
                res = re.match('%+', match)
                if len(res.group(0)) % 2:
                    self.output.add_info('W', self.pkg, 'macro-in-comment', match)

    def _checkline_python_setup_test(self, line):
        # Test if the "python setup.py test" deprecated subcommand is used
        if self.current_section == 'check' and python_setup_test_regex.search(line):
            self.output.add_info('W', self.pkg, 'python-setup-test', line[:-1])

    def _checkline_python_setup_install(self, line):
        # Test if the "python setup.py install" deprecated subcommand is used
        if self.current_section == 'install' and python_setup_install_regex.search(line):
            self.output.add_info('W', self.pkg, 'python-setup-install', line[:-1])

    def _checkline_python_module_def(self, line):
        """
        Test if the "python_module" macro is defined in the spec file
        This macro was in py2pack but now it should be provided by
        python-rpm-macros
        """
        if python_module_def_regex.search(line):
            self.output.add_info('W', self.pkg, 'python-module-def', line[:-1])

    def _checkline_python_sitelib_glob(self, line):
        """Test if %{python_sitelib}/* is present in %files section."""
        if self.current_section != 'files':
            return

        if python_sitelib_glob_regex.match(line):
            self.output.add_info('W', self.pkg, 'python-sitelib-glob-in-files',
                                 line[:-1])

    def _checkline_shared_dir_glob(self, line):
        """Test if %{_bindir}/*, etc. is present in %files section."""
        if self.current_section != 'files':
            return

        if shared_dir_glob_regex.match(line):
            self.output.add_info('W', self.pkg, 'shared-dir-glob-in-files',
                                 line[:-1])

    def _checkline_forbidden_controlchars(self, line):
        """Look for controlchar in any line"""
        # https://github.com/rpm-software-management/rpmlint/issues/1067
        if Pkg.has_forbidden_controlchars(line):
            self.output.add_info('W', self.pkg, 'forbidden-controlchar-found')

    def _check_suse_update_desktop_file(self, line):
        """
        Test if update-desktop-files is in BuildRequires. The usage of
        %suse_update_desktop_file is deprecated now.
        """
        if suse_update_desktop_file_regex.match(line):
            # Don't show the message for yast, there's no migration path yet.
            if 'yast' in self.pkg.name.lower():
                return

            self.output.add_info('W', self.pkg,
                                 'suse-update-desktop-file-deprecated',
                                 '%suse_update_desktop_file is deprecated')



from importlib import metadata
from pathlib import Path
import platform
import re

from packaging.requirements import InvalidRequirement, Requirement
from rpmlint.checks.AbstractCheck import AbstractFilesCheck

# Warning messages
WARNS = {
    'doc': 'python-doc-in-package',
    'sphinx': 'python-sphinx-doctrees-leftover',
}

# Error messages
ERRS = {
    'egg-distutils': 'python-egg-info-distutils-style',
    'tests': 'python-tests-in-site-packages',
    'doc': 'python-doc-in-site-packages',
    'src': 'python-src-in-site-packages',
}

SITELIB_RE = '/usr/lib[^/]*/python([^/]*)/site-packages'

# Paths that shouldn't be in any packages, ever, because they clobber global
# name space.
ERR_PATHS = [
    (re.compile(f'{SITELIB_RE}/tests?$'), 'tests'),
    (re.compile(f'{SITELIB_RE}/docs?$'), 'doc'),
    (re.compile(f'{SITELIB_RE}/src$'), 'src'),
]

# Paths that shouldn't be in any packages, but might need to be under
# sufficiently special circumstances.
WARN_PATHS = [
    (re.compile(f'{SITELIB_RE}/[^/]+/docs?$'), 'doc'),
    (re.compile(r'.*/\.doctrees$'), 'sphinx'),
]


EGG_INFO_RE = re.compile('.*egg-info$')
PYC_RE = re.compile(r'cpython-(\d+)')


class PythonCheck(AbstractFilesCheck):
    def __init__(self, config, output):
        super().__init__(config, output, r'.*')

    def check_binary(self, pkg):
        self._pyc_version = None
        super().check_binary(pkg)

    def check_file(self, pkg, filename):
        # egg-info format
        is_egginfo = filename.endswith('egg-info/requires.txt')
        # dist-info format
        is_distinfo = filename.endswith('dist-info/METADATA')
        if is_egginfo or is_distinfo:
            self._check_requires(pkg, filename)
            return

        if EGG_INFO_RE.match(filename):
            self._check_egginfo(pkg, filename)
            return

        for path_re, key in WARN_PATHS:
            if path_re.match(filename):
                if key == 'doc':
                    # Check for __init__.py file inside doc, maybe this is a
                    # module, not documentation
                    module_file = f'{filename}/__init__.py'
                    if module_file in pkg.files.keys():
                        continue
                self.output.add_info('W', pkg, WARNS[key], filename)

        for path_re, key in ERR_PATHS:
            if path_re.match(filename):
                self.output.add_info('E', pkg, ERRS[key], filename)

        if filename.endswith('pyc'):
            # Check for .pyc from different python versions in the same package
            self._check_multiple_python_pyc(pkg, filename)

    def _check_multiple_python_pyc(self, pkg, filename):
        """
        """
        search = PYC_RE.search(filename)
        if not search:
            return

        version = search.group(1)
        if self._pyc_version is None:
            # First pyc file, just store the version
            self._pyc_version = version
        elif self._pyc_version != version:
            self.output.add_info('W', pkg, 'python-pyc-multiple-versions',
                                 'expected:', self._pyc_version, filename)

    def _check_egginfo(self, pkg, filename):
        """
        Check type of egg-info metadata and check Requires against egg-info
        metadata if applicable.
        """

        filepath = Path(pkg.dir_name() or '/', filename.lstrip('/'))
        # Check for (deprecated) distutils style metadata.
        if filepath.is_file():
            self.output.add_info('E', pkg, ERRS['egg-distutils'], filename)

    def _check_requires(self, pkg, filename):
        """
        Look for all requirements defined in the python package and
        compare with the requirements defined in the rpm package
        """

        filepath = Path(pkg.dir_name() or '/', filename.lstrip('/'))
        d = metadata.PathDistribution.at(filepath.parent)
        if not d.requires:
            return

        requirements = []
        for requirement in d.requires:
            # Ignore broken requirements
            try:
                req = Requirement(requirement)
            except InvalidRequirement:
                continue

            requirements.append(req)

        self._check_requirements(pkg, requirements, d)

    def _check_requirements(self, pkg, requirements, distribution):
        """
        Check mismatch between the list of requirements and the rpm
        declared requires.
        """

        env = {
            'python_version': '.'.join(platform.python_version_tuple()[:2]),
            'os_name': 'posix',
            'platform_system': 'Linux',
        }

        # Look for python version
        for req in pkg.requires:
            if req.name == 'python(abi)':
                _, pyv, _ = req.version
                env['python_version'] = pyv
                break

        # python_version from distribution path
        python_path = re.findall(SITELIB_RE, str(distribution._path))
        if python_path:
            env['python_version'] = python_path[0]

        # Check for missing requirements
        for req in requirements:
            if req.marker:
                # Ignore extra requires
                if 'extra' in str(req.marker):
                    continue
                # Ignore not env requirements
                if not req.marker.evaluate(environment=env):
                    continue

            self._check_require(pkg, req)

        # Check for python requirement not needed
        self._check_leftover_requirements(pkg, requirements, env)

    def _check_require(self, pkg, requirement):
        """
        Look for the module_name in the package requirements, looking
        for common python rpm package names like python-foo,
        python3-foo, etc.
        """

        names = self._module_names(requirement.name, extras=requirement.extras)

        # Add pythonX-foo variants
        names += [f'python\\d*-{re.escape(i)}' for i in names]
        # Add python3.12dist(foo) variants
        names += [f'python\\d+(\\.\\d+)?dist\\({re.escape(i)}\\)' for i in names]
        regex = '|'.join(names)
        # Support complex requirements like
        # (python310-jupyter-server >= 1.15 with python310-jupyter-server < 3)
        version_req = r'\s*(==|<|<=|>|>=)\s*[\w.]+\s*'
        richop_req = r'\s+(and|or|if|unless|else|with|without)\s+.*'
        try:
            regex = re.compile(rf'^\(?({regex})({version_req})?({richop_req})?\)?\s*$', re.IGNORECASE)
        except re.error:
            # Bad regular expression, it could be a name with weird
            # characters
            return False

        for req in pkg.req_names:
            if regex.match(req):
                return True

        self.output.add_info('W', pkg, 'python-missing-require', requirement.name)
        return False

    def _check_leftover_requirements(self, pkg, requirements, env):
        """
        Look for python-foo requirements in the rpm package that are
        not in the list of requirements of this package.
        """

        pythonpac = re.compile(r'^python\d*-(?P<name>.+)$')
        reqs = set()
        for i in requirements:
            # Ignore not env requirements
            if i.marker and not i.marker.evaluate(environment=env):
                continue
            for n in self._module_names(i.name, extras=i.extras):
                reqs.add(n.lower())

        for req in pkg.req_names:
            match = pythonpac.match(req)
            if not match:
                continue

            module_name = match.group('name').strip().lower()
            # ignore python-base, python-devel packages, that are
            # not python modules
            if module_name in ('base', 'devel'):
                continue

            names = set(self._module_names(module_name))

            if not (names & reqs):
                self.output.add_info('W', pkg, 'python-leftover-require', req)

    def _module_names(self, module_name, extras=None):
        """
        Return a list with possible variants of the module name,
        replacing "-", "_".
        """

        # Name variants changing '-' with '_'
        variants = []
        variants.append(module_name.replace('-', '_'))
        variants.append(module_name.replace('_', '-'))

        # Look also for python-MOD-EXTRA
        if extras:
            for e in extras:
                variants += self._module_names(f'{module_name}-{e}')

        return [
            module_name,
            *variants,
        ]



from mockdata.mock_tags import (
    FooDevelPackage,
    FuseCommonPackage,
    InvalidExceptionPackage,
    MissingProvidesPackage,
    SelfPackage,
    UnexpandedMacroPackage,
    ValidExceptionPackage,
)
import pytest
from rpmlint.checks.TagsCheck import TagsCheck
from rpmlint.filter import Filter
import rpmlint.spellcheck

from Testing import CONFIG, get_tested_package
from Testing import get_tested_mock_package
from Testing import HAS_ENGLISH_DICTIONARY, HAS_FRENCH_DICTIONARY


@pytest.fixture(scope='function', autouse=True)
def tagscheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    yield output, test


@pytest.fixture
def output(tagscheck):
    output, _test = tagscheck
    yield output


@pytest.fixture
def test(tagscheck):
    _output, test = tagscheck
    yield test


@pytest.mark.parametrize('package', [UnexpandedMacroPackage])
def test_unexpanded_macros(package, tagscheck):
    output, test = tagscheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'unexpanded-macro Recommends' in out
    assert 'unexpanded-macro Provides' in out
    assert 'unexpanded-macro Conflicts' in out
    assert 'unexpanded-macro Suggests' in out
    assert 'unexpanded-macro Obsoletes' in out
    assert 'unexpanded-macro Enhances' in out


@pytest.mark.parametrize('package', [SelfPackage])
def test_self_provides(package, tagscheck):
    output, test = tagscheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: useless-provides self' in out


@pytest.mark.parametrize('package', [FuseCommonPackage])
def test_useless_provides_only_versions(package, tagscheck):
    output, test = tagscheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: useless-provides fuse-common' not in out


@pytest.mark.parametrize('package', [FooDevelPackage])
def test_development_package(package, tagscheck):
    output, test = tagscheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'W: devel-package-with-non-devel-group Games' in out


@pytest.mark.parametrize('package', [MissingProvidesPackage])
def test_missing_provides(package, tagscheck):
    output, test = tagscheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: no-pkg-config-provides' in out


@pytest.mark.parametrize('package', [InvalidExceptionPackage])
def test_invalid_license_exception(package, tagscheck):
    output, test = tagscheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'W: invalid-license-exception sparta' in out


@pytest.mark.parametrize('package', [ValidExceptionPackage])
def test_valid_license_exception(tmp_path, package, tagscheck):
    CONFIG.info = True
    CONFIG.configuration['ValidLicenseExceptions'] = ['389-exception']
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(package)
    out = output.print_results(output.results)
    assert 'W: invalid-license-exception' not in out


@pytest.mark.parametrize('package', ['source/valid-exception-in-grouping'])
def test_valid_license_exception_in_grouping(tmp_path, package, tagscheck):
    CONFIG.info = True
    CONFIG.configuration['ValidLicenses'] = ['BSD-3-Clause', 'GPL-2.0-only']
    CONFIG.configuration['ValidLicenseExceptions'] = ['Qt-GPL-exception-1.0']
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'W: invalid-license-exception' not in out


@pytest.mark.parametrize('package', ['source/valid-exception-begin-grouping'])
def test_valid_license_exception_begin_grouping(tmp_path, package, tagscheck):
    CONFIG.info = True
    CONFIG.configuration['ValidLicenses'] = ['BSD-3-Clause', 'GPL-2.0-only']
    CONFIG.configuration['ValidLicenseExceptions'] = ['Qt-GPL-exception-1.0']
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'W: invalid-license-exception' not in out


@pytest.mark.parametrize('package', ['binary/xtables-addons-kmp-default'])
def test_forbidden_controlchar_found_requires(tmp_path, package, tagscheck):
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: forbidden-controlchar-found Requires:' in out


@pytest.mark.parametrize('package', ['binary/ruby2.6-rubygem-fast_gettext'])
def test_forbidden_controlchar_found_changelog(tmp_path, package, tagscheck):
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: forbidden-controlchar-found %changelog' in out


@pytest.mark.parametrize('package', ['binary/SpecCheck4'])
def test_forbidden_controlchar_found(tmp_path, package, tagscheck):
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: forbidden-controlchar-found Requires:' in out
    assert 'E: forbidden-controlchar-found Provides:' in out
    assert 'E: forbidden-controlchar-found Obsoletes:' in out
    assert 'E: forbidden-controlchar-found Conflicts:' in out
    assert 'E: forbidden-controlchar-found %changelog :' in out


@pytest.mark.parametrize('package', ['binary/unexpanded-macro-exp'])
def test_check_unexpanded_macro(tmp_path, package, tagscheck):
    """Test if a package has an unexpanded macro in it's specfile."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'W: unexpanded-macro Packager %ppc' in out
    assert 'W: unexpanded-macro Group %ppc' in out
    assert 'W: unexpanded-macro Provides /something%ppc %ppc' in out
    assert 'W: unexpanded-macro Conflicts something:%ppc %ppc' in out
    assert 'W: unexpanded-macro Supplements packageand(python-gobject:%{gdk_real_package})%ppc %ppc' in out
    assert 'W: unexpanded-macro Suggests /%ppc %ppc' in out
    assert 'W: unexpanded-macro Enhances /%ppc %ppc' in out
    assert 'W: unexpanded-macro Recommends /%ppc %ppc' in out
    assert 'W: unexpanded-macro Supplements packageand(python-gobject:%{gdk_real_package})%ppc %{gdk_real_package}' in out
    assert 'W: unexpanded-macro Supplements packageand(python-gobject:%{gdk_real_package})%ppc %ppc' in out


@pytest.mark.parametrize('package', ['binary/invalid-version'])
def test_check_errors(tmp_path, package, tagscheck):
    """Test package for check invalid-version."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package has a Version: tag with pre/alpha/beta suffixes in it's specfile
    assert 'E: invalid-version 0pre' in out


# TODO: Add a test for no-changelogname-tag not in out


@pytest.mark.parametrize('package', ['binary/summary-warning'])
def test_check_summary_warning(tmp_path, package, tagscheck):
    """Test package for check
    - in out,
        summary-too-long, summary-has-leading-spaces,
        description-shorter-than-summary,
    - not in out
    invalid-version, unexpanded-macro.
    """
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if package has a summary longer than 80 characters
    assert 'E: summary-too-long' in out
    # Test if package has leading space at the beginning of the summary
    # where non-breaking space is used (U+00A0).
    assert b'E: summary-has-leading-spaces \xc2\xa0\xc2\xa0lorem'.decode() in out
    # Test if package has a shorter description than Summary
    assert 'W: description-shorter-than-summary' in out
    # Test if a package has a Version: tag
    # without any pre/alpha/beta suffixes in it's specfile.
    assert 'E: invalid-version' not in out
    # Test if a package does not have an unexpanded
    # macro in it's specfile.
    assert 'W: unexpanded-macro' not in out
    assert 'E: no-group-tag' in out


@pytest.mark.parametrize('package', ['binary/no-url-tag'])
def test_check_warning(tmp_path, package, tagscheck):
    """Test if a package contains the warning for
    summary-not-capitalized, summary-ended-with-dot,
    no-url-tag."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if package Summary does not start with a capital letter
    assert 'W: summary-not-capitalized no-url-tag warning.' in out
    # Test if a package Summary ends with a period
    assert 'W: summary-ended-with-dot no-url-tag warning.' in out
    # Test if a package does not contain a Url: tag
    assert 'W: no-url-tag' in out


@pytest.mark.parametrize('package', ['binary/invalid-la-file'])
def test_check_errors_not_found(tmp_path, package, tagscheck):
    """Test packages for checks
    summary-too-long, summary-not-capitalized,
    summary-ended-with-dot, summary-has-leading-spaces,
    no-url-tag, description-shorter-than-summary.
    """
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package Summary is not longer than 80 characters
    assert 'W: summary-too-long' not in out
    # Test if Summary starts with a capitalized word
    assert 'W: summary-not-capitalized' not in out
    # Test if a package Summary does not end with a period
    assert 'W: summary-ended-with-dot' not in out
    # Test if a package Summary does not begin with a leading space
    assert 'W: summary-has-leading-spaces' not in out
    # Test if a package contains a Url: tag
    assert 'W: no-url-tag' not in out
    # Test if package no shorter description than Summary
    assert 'W: description-shorter-than-summary' not in out


@pytest.mark.parametrize('package', ['binary/misc-warnings'])
def test_check_misc_warning(tmp_path, package, tagscheck):
    """Test package for check tag-in-description,
    name-repeated-in-summary, invalid-url."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package has a tag such as Name: in the description
    assert 'W: tag-in-description Name:' in out
    # Test if a package has it's name repeated in the summary tag
    assert 'W: name-repeated-in-summary misc-warnings' in out
    # Test if a package has a invalid url value in the URL: tag
    assert 'W: invalid-url URL so;mething.' in out


@pytest.mark.parametrize('package', ['binary/misc-no-warnings'])
def test_check_misc_warning_not_found(tmp_path, package, tagscheck):
    """Test package for check not in out
    tag-in-description, name-repeated-in-summary, invalid-url."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package does not have a tag in description
    assert 'W: tag-in-description' not in out
    # Test if a package does not have a name repeated in the summary tag value
    assert 'W: name-repeated-in-summary' not in out
    # Test if a package does not have a invalid url in the URL tag value
    assert 'W: invalid-url' not in out


@pytest.mark.parametrize('package', ['binary/invalid-dependency'])
def test_check_invalid_dependency(tmp_path, package, tagscheck):
    """Test if a package has
    invalid-dependency, no-description-tag, unreasonable-epoch."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package has a Epoch tag value greater than 99
    assert 'W: unreasonable-epoch 100' in out
    # Test if a package has Requires: starts with /usr/local
    assert 'E: invalid-dependency /usr/local/something' in out
    # Test if a package has empty or no description tag
    assert 'E: no-description-tag' in out


@pytest.mark.parametrize('package', ['binary/random-exp'])
def test_package_random_warnings(tmp_path, package, tagscheck):
    """Test if a package has check,
    - in out,
        obsolete-not-provided
        description-line-too-long,
    - not in out,
        invalid-dependency,
        unreasonable-epoch,
        no-description-tag,
        self-obsoletion."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package that was obsoleted is still provided
    # in newer package to avoid unnecessary dependency breakage
    assert 'W: obsolete-not-provided' in out
    # Test if a package has a description longer than 79 characters
    assert 'E: description-line-too-long This is ridiculously long description that has no meaning but is used to test the check description-line-too-long.' in out
    # Test if a package has a Requires tag value start with
    # string other than /usr/local/
    assert 'E: invalid-dependency' not in out
    # Test if a package has a Epoch tag value less than or equal to 99
    assert 'W: unreasonable-epoch' not in out
    # Test if a package has a non-empty description tag
    assert 'W: no-description-tag' not in out
    # Test if a package does not have itself in Obsoletes: tag value
    assert 'W: self-obsoletion' not in out


@pytest.mark.parametrize('package', ['binary/random-devel'])
def test_package_random_exp(tmp_path, package, tagscheck):
    """Test if a package check,
    - in out,
        self-obsoletion,
    - not in out,
        obsolete-not-provided,
        description-line-too-long,
        devel-dependency."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package obsoletes itself i.e. Obsoletes: random-devel
    assert 'W: self-obsoletion random-devel obsoletes random-devel' in out
    # Test if a obsoleted package has been provided to
    # to avoid unnecessary breakage i.e. Requires: python
    assert 'W: obsolete-not-provided' not in out
    # Test if a package has a description line not greater 79 characters
    assert 'E: description-line-too-long' not in out
    # Test if a package is a *-devel package and requires a devel dependency
    assert 'W: devel-dependency' not in out


@pytest.mark.parametrize('package', ['binary/requires-on-release'])
def test_check_requires_on_release(tmp_path, package, tagscheck):
    """Test if a package check,
    - in out,
        requires-on-release."""
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package requires specific release of another package
    assert 'W: requires-on-release baz = 2.1-1' in out


@pytest.mark.parametrize('package', ['binary/invalid-license'])
def test_check_invalid_license(tmp_path, package, tagscheck):
    """Test if a package check,
    - in out,
        invalid-license,
    - not in out,
        requires-on-release."""
    CONFIG.configuration['ValidLicenses'] = ['MIT']
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package has a License: tag value different from
    # ValidLicense = [] list in configuration
    assert 'W: invalid-license Apache License' in out
    # Test if a package does not Requires: a specific version of a package
    assert 'W: requires-on-release' not in out


@pytest.mark.parametrize('package', ['binary/not-standard-release-extension'])
def test_package_not_std_release_extension(tmp_path, package, tagscheck):
    """Test if package has check,
    - in out,
        not-standard-release-extension
    - not in out,
        invalid-license."""
    CONFIG.configuration['ReleaseExtension'] = 'hello$'
    CONFIG.configuration['ValidLicenses'] = ['Apache-2.0 License']
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package has a ReleaseExtension regex does not match with the Release: tag value expression
    # i.e. Release tag value must not match regex expression 'hello$'
    assert 'W: not-standard-release-extension 1.1' in out
    # Test if a package does have the same License value as defined in the ValidLicense in configdefaults
    assert 'W: invalid-license Apache-2.0 License' not in out


@pytest.mark.parametrize('package', ['binary/non-standard-group'])
def test_check_non_standard_group(tmp_path, package, tagscheck):
    """Test if a package has check,
    - in out,
        non-standard-group
    - not in out,
        not-standard-release-extension."""
    CONFIG.configuration['ValidGroups'] = ['Devel/Something']
    CONFIG.configuration['ReleaseExtension'] = '0'
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package has a different Group: tag value than ValidGroups = []
    assert 'W: non-standard-group non/standard/group' in out
    # Test if a package matches the Release tag regex
    assert 'not-standard-release-extension 0' not in out


@pytest.mark.parametrize('package', ['binary/dev-dependency'])
def test_package_dev_dependency(tmp_path, package, tagscheck):
    """Test if a package check,
    - in out,
        devel-dependency,
    - not in out,
        non-standard-group."""
    CONFIG.configuration['ValidGroups'] = ['Devel/Something']
    output = Filter(CONFIG)
    test = TagsCheck(CONFIG, output)
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # Test if a package is not a devel package itself but requires a devel dependency
    assert 'E: devel-dependency glibc-devel' in out
    # Test if a package does not have a Group tag
    assert 'W: non-standard-group Devel/Something' not in out


@pytest.mark.parametrize('package', ['binary/summary-on-multiple-lines'])
def test_summary_on_multiple_lines(tmp_path, package, tagscheck):
    # Test if a package has summary on multiple lines.
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: summary-on-multiple-lines' in out


@pytest.mark.skipif(not rpmlint.spellcheck.ENCHANT, reason='Missing enchant bindings')
@pytest.mark.skipif(not HAS_ENGLISH_DICTIONARY, reason='Missing English dictionary')
@pytest.mark.skipif(not HAS_FRENCH_DICTIONARY, reason='Missing French dictionary')
@pytest.mark.parametrize('package', [
    'binary/spellingerrors-default',
    'binary/spellingerrors-lang',
    'binary/spellingerrors-lang2',
    'binary/spellingerrors-lang3',
])
def test_description_spelling_error(tmp_path, package, tagscheck):
    output, test = tagscheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: spelling-error' in out


@pytest.mark.parametrize('package', [
    get_tested_package('binary/xrootd-devel', '/tmp/'),
    get_tested_mock_package(
        name='xrootd-devel',
        files=[
            '/usr/lib64/libXrdXml.so',
        ],
        header={
            'requires': [
                'xrootd-libs(x86-64) = 1:5.6.3-2.fc39',
            ],
            'ARCH': 'noarch',
            'NAME': 'xrootd-devel',
            'VERSION': '5.6.3',
            'RELEASE': '2.fc39',
            'EPOCH': 1,
        },
    ),
])
def test_missing_dependency_on_with_epoch(package, output, test):
    test.check(package)
    out = output.print_results(output.results)
    assert 'W: missing-dependency-on' not in out
